/*
 * ConversionHandler.cpp
 *
 *  Created on: 23. 2. 2014
 *	  Author: tomas
 */

#include "ConversionHandler.h"

#include "conversions/fa2re/unbounded/StateEliminationUnbounded.h"
#include "conversions/fa2re/formal/StateEliminationFormal.h"
#include "conversions/fa2re/Algebraic.h"

#include "conversions/re2fa/GlushkovNFA.h"
#include "conversions/re2fa/Thompson.h"
#include "conversions/re2fa/BrzozowskiDerivation.h"

#include "conversions/fa2rg/fa2lrg/FAtoLRGConverter.h"
#include "conversions/fa2rg/fa2rrg/FAtoRRGConverter.h"

#include "conversions/rg2fa/RGtoFA.h"

#include "conversions/rg2re/Algebraic.h"

#include "conversions/re2rg/re2rrg/GlushkovNFA.h"
#include "conversions/re2rg/re2rrg/BrzozowskiDerivation.h"

#include "conversions/rg2rg/LeftToRightRegularGrammar.h"
#include "conversions/rg2rg/RightToLeftRegularGrammar.h"


using namespace alib;
using namespace automaton;
using namespace grammar;
using namespace regexp;
using namespace sax;
using namespace std;

ConversionHandler::ConversionHandler( list<Token> & tokens, const std::string & target, const std::string & algorithm, std::ostream & out ) :
	m_tokens( tokens ),
	m_source( parseFormalismFromTokens( ) ),
	m_target( parseFormalismFromString( target ) ),
	m_algorithm( parseAlgorithmFromString( algorithm ) ),
	m_out( out )
{

}

void ConversionHandler::convert( void )
{
	if( m_source == FINITE_AUTOMATON )
		convertFSM( );
	else if( isRegExp(m_source) )
		convertRE( );
	else if( isGrammar( m_source ) )
		convertRG( );
	else
		throw exception::AlibException( "ConversionHandler: Unknown source formalism." );
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertFSM( void )
{
	if( m_target == FINITE_AUTOMATON )
		throw exception::AlibException( "ConversionHandler: No valid conversion from FSM to FSM." );
	else if( isRegExp( m_target ) )
		convertFSMtoRE( );
	else if( isGrammar( m_target ) )
		convertFSMtoRG( );
	else
		throw exception::AlibException( "ConversionHandler: Unknown target formalism." );
}

void ConversionHandler::convertRE( void )
{
	if( m_target == FINITE_AUTOMATON )
		convertREtoFSM( );
	else if( isRegExp( m_target ) )
		throw exception::AlibException( "ConversionHandler: No valid conversion from RE to RE." );
	else if( isGrammar( m_target ) )
		convertREtoRG( );
	else
		throw exception::AlibException( "ConversionHandler: Unknown target formalism." );
}

void ConversionHandler::convertRG( void )
{
	if( m_target == FINITE_AUTOMATON )
		convertRGtoFSM( );
	else if( isRegExp( m_target ) )
		convertRGtoRE( );
	else if( isGrammar( m_target ) )
		convertRGtoRG( );
	else
		throw exception::AlibException( "ConversionHandler: Unknown target formalism." );
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertFSMtoRE(void)
{
    switch( m_target )
    {
    case REGULAR_EXPRESSION_FORMAL:
        convertFSMtoREF();
        break;
    default:
        convertFSMtoREU();
        break;
    }
}

void ConversionHandler::convertFSMtoREF(void)
{
	const automaton::Automaton automaton = alib::DataFactory::fromTokens<automaton::Automaton>(m_tokens);

	switch( m_algorithm )
	{
	case BRZOZOWSKI_ALGEBRAIC:
        break;
	case STATE_ELIMINATION:
	default: {
			alib::DataFactory::toStdout(conversions::fa2re::StateEliminationFormal::convert(automaton));
			break;
		}
	}

}

void ConversionHandler::convertFSMtoREU( void )
{
	const automaton::Automaton automaton = alib::DataFactory::fromTokens<automaton::Automaton>(m_tokens);

	switch( m_algorithm )
	{
	case BRZOZOWSKI_ALGEBRAIC: {
			std::string xmlMark = m_tokens.front( ).getData( );
			if( xmlMark == "NFA") {
				const automaton::NFA fsm = alib::DataFactory::fromTokens<automaton::NFA>( m_tokens );
				regexp::UnboundedRegExp re = conversions::fa2re::Algebraic::convert( fsm );
				alib::DataFactory::toStdout(re);
			} else if( xmlMark == "DFA") {
				const automaton::DFA fsm = alib::DataFactory::fromTokens<automaton::DFA>( m_tokens );
				regexp::UnboundedRegExp re = conversions::fa2re::Algebraic::convert( fsm );
				alib::DataFactory::toStdout(re);
			} else {
				throw exception::AlibException("Unrecognised formalism");
			}
			break;
		}
	case STATE_ELIMINATION:
	default: {
			alib::DataFactory::toStdout(conversions::fa2re::StateEliminationUnbounded::convert(automaton));
			break;
		}
	}
}

void ConversionHandler::convertFSMtoRG( void )
{
	if( m_target == RIGHT_REGULAR_GRAMMAR )
		convertFSMtoRRG( );
	else if( m_target == LEFT_REGULAR_GRAMMAR )
		convertFSMtoLRG( );
	else
		throw exception::AlibException( "ConversionHandler::Unknown target formalism" );
}

void ConversionHandler::convertFSMtoRRG( void )
{
	const automaton::Automaton fsm = alib::DataFactory::fromTokens<automaton::Automaton>( m_tokens );

	switch( m_algorithm )
	{
	default: {
			conversions::fa2rg::FAtoRRGConverter conv;
			alib::DataFactory::toStdout(conv.convert(fsm));
			break;
		}
	}
}

void ConversionHandler::convertFSMtoLRG( void )
{
	const automaton::Automaton fsm = alib::DataFactory::fromTokens<automaton::Automaton>( m_tokens );

	switch( m_algorithm )
	{
	default: {
			conversions::fa2rg::FAtoLRGConverter conv;
			alib::DataFactory::toStdout(conv.convert(fsm));
			break;
		}
	}
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertREtoFSM( void )
{
    const regexp::RegExp regexp = alib::DataFactory::fromTokens<regexp::RegExp>(m_tokens);

	switch( m_algorithm )
	{
	case BRZOZOWSKI_DERIVATION: {
			alib::DataFactory::toStdout(conversions::re2fa::BrzozowskiDerivation::convert(regexp));
			break;
		}
	case THOMPSON_NFA: {
			alib::DataFactory::toStdout(conversions::re2fa::Thompson::convert(regexp));
			break;
		}
	case GLUSHKOV_NFA:
	default: {
			alib::DataFactory::toStdout(conversions::re2fa::GlushkovNFA::convert(regexp));
			break;
		}
	}
}

void ConversionHandler::convertREtoRG( void )
{
	if( m_target == RIGHT_REGULAR_GRAMMAR )
		convertREtoRRG( );
	else if( m_target == LEFT_REGULAR_GRAMMAR )
		throw exception::AlibException( "ConversionHandler:: RE to LRG is not implemented. Please convert to RRG and then to LRG." );
	else
		throw exception::AlibException( "ConversionHandler::Unknown target formalism" );
}

void ConversionHandler::convertREtoRRG( void )
{
	const regexp::RegExp regexp = alib::DataFactory::fromTokens<regexp::RegExp>(m_tokens);

	switch(m_algorithm)
	{
	case BRZOZOWSKI_DERIVATION: {
			alib::DataFactory::toStdout(conversions::re2rg::BrzozowskiDerivation::convert(regexp));
			break;
		}
	default: {
			alib::DataFactory::toStdout(conversions::re2rg::GlushkovNFA::convert(regexp));
			break;
		}
	}
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertRGtoFSM( void )
{
	const grammar::Grammar grammar = alib::DataFactory::fromTokens<grammar::Grammar>(m_tokens);
    alib::DataFactory::toStdout(conversions::rg2fa::RGtoFA::convert(grammar));
}

void ConversionHandler::convertRGtoRG( void )
{
	if( m_source == LEFT_REGULAR_GRAMMAR )
		convertLRGtoRRG( );
	else if( m_source == RIGHT_REGULAR_GRAMMAR )
		convertRRGtoLRG( );
	else if( m_source == LEFT_LINEAR_GRAMMAR )
		throw exception::AlibException( "ConversionHandler: Conversions from linear grammars are not yet implemented." );
	else if( m_source == RIGHT_LINEAR_GRAMMAR )
		throw exception::AlibException( "ConversionHandler: Conversions from linear grammars are not yet implemented." );
	else
		throw exception::AlibException( "ConversionHandler: Unknown target formalism." );
}

void ConversionHandler::convertRGtoRE( void )
{
	const grammar::Grammar grammar = alib::DataFactory::fromTokens<grammar::Grammar>(m_tokens);

	switch( m_algorithm )
	{
	case BRZOZOWSKI_ALGEBRAIC:
	default:
		alib::DataFactory::toStdout(conversions::rg2re::Algebraic::convert(grammar));
        break;
	}
}

void ConversionHandler::convertLRGtoRRG( void )
{
	const grammar::LeftRG lrg = alib::DataFactory::fromTokens<grammar::LeftRG>( m_tokens );

	switch( m_algorithm )
	{
	default:
		alib::DataFactory::toStdout(conversions::rg2rg::LeftToRightRegularGrammar::convert(lrg));
		break;
	}
}

void ConversionHandler::convertRRGtoLRG( void )
{
	const grammar::RightRG rrg = alib::DataFactory::fromTokens<grammar::RightRG>( m_tokens );

	switch( m_algorithm )
	{
	default:
		alib::DataFactory::toStdout(conversions::rg2rg::RightToLeftRegularGrammar::convert(rrg));
		break;
	}
}

// ----------------------------------------------------------------------------

ConversionHandler::TFormalism ConversionHandler::parseFormalismFromString( const std::string & _target ) const
{
	std::string target;
	for( const auto & c : _target )
		target.append( 1, tolower( c ) );

	if( target == "fa" || target == "fsa" || target == "fsm" )
		return FINITE_AUTOMATON;

	if( target == "re" || target == "regexp" || target == "regex" )
		return REGULAR_EXPRESSION_UNBOUNDED;
    if (target == "formalre" || target == "formalregexp" || target == "formalregex")
        return REGULAR_EXPRESSION_FORMAL;

	if( target == "rrg" || target == "rg" || target == "grammar" )
		return RIGHT_REGULAR_GRAMMAR;

	if( target == "lrg" )
		return LEFT_REGULAR_GRAMMAR;

	throw exception::AlibException( "ConversionHandler::Unknown target type" );
}

ConversionHandler::TFormalism ConversionHandler::parseFormalismFromTokens( void ) const
{
	std::string xmlMark = m_tokens.front( ).getData( );

	if( xmlMark == "automaton" || xmlMark == "NFA" || xmlMark == "DFA" )
		return FINITE_AUTOMATON;

	if( xmlMark == "regexp" || xmlMark == "unboundedRegexp" )
		return REGULAR_EXPRESSION_UNBOUNDED;

    if(xmlMark == "formalRegExp")
        return REGULAR_EXPRESSION_FORMAL;

	if( xmlMark == "RightRG" )
		return RIGHT_REGULAR_GRAMMAR;

    if( xmlMark == "LeftRG" )
        return LEFT_REGULAR_GRAMMAR;

	throw exception::AlibException( "ConversionHandler: Invalid input formalism." );
}

ConversionHandler::TAlgorithm ConversionHandler::parseAlgorithmFromString( const std::string & _algorithm ) const
{
	std::string algorithm;
	for( const auto & c : _algorithm )
		algorithm.append( 1, tolower( c ) );

	if( algorithm == "elimination" ) return STATE_ELIMINATION;
	if( algorithm == "algebraic" ) return BRZOZOWSKI_ALGEBRAIC;
	if( algorithm == "brzozowski" ) return BRZOZOWSKI_DERIVATION;
	if( algorithm == "glushkov" ) return GLUSHKOV_NFA;
	if( algorithm == "thompson" ) return THOMPSON_NFA;

	if( algorithm == "" ) return DEFAULT;

	throw exception::AlibException( "ConversionHandler:: Invalid algorithm. See help." );
}
