#include <iostream>
#include <getopt.h>
#include <string.h>
#include <ctype.h>

#include "automaton/Automaton.h"
#include "automaton/FSM/NFA.h"
#include "automaton/FSM/DFA.h"
#include "automaton/PDA/NPDA.h"
#include "factory/DataFactory.hpp"
#include "exception/AlibException.h"

#include "determinize/nfa/NFADeterminizer.h"
#include "determinize/idpda/IDPDADeterminizer.h"
#include "determinize/vpa/VPADeterminizer.h"
#include "determinize/hdpda/HDPDADeterminizer.h"
#include "automaton/PDAToRHPDA.h"
#include "automaton/RHPDAToPDA.h"

#define VERSION "0.0.1"

#define TYPE_FSM "fsm"
#define TYPE_IDPDA "idpda"
#define TYPE_VPA "vpa"
#define TYPE_RHDPDA "rhdpda"

/**
 * Prints help to standard output.
 */
void printHelp() {
	std::cout << "adeterminize " << VERSION << std::endl;
	std::cout << "Determinize various types of automaton." << std::endl;
	std::cout << "Usage: adeterminize -t TYPE [SWITCH...]" << std::endl;
	std::cout << "Possible arguments:" << std::endl;
	std::cout << "  -t, --type=TYPE \t Specifies type of input automaton, possible values are 'FSM' for final-state machine, 'IDPDA' for input-driven pushdown automaton, 'VPA' for visible pushdown automaton and 'RHDPDA' for real-time height deterministic pushdown automaton." << std::endl;
	std::cout << "  -h, --help \t Displays this help message." << std::endl;
	std::cout << std::endl;
}


/**
 * Converts given type to lower case.
 *
 * @param type
 * @return type in lower case
 */
std::string getLowerCaseType(std::string type) {
	std::string lowerCaseType;
	for (const auto& c : type) {
		lowerCaseType.append(1, tolower(c));
	}
	return lowerCaseType;
}


int main(int argc, char** argv) {
	static struct option longOptions[] = {
	  {"type", required_argument, NULL, 't'},
	  {"algorithm", required_argument, NULL, 'a'},
	  {"version", no_argument, NULL, 'v'},
	  {"help", no_argument, NULL, 'h'},
	  {0, 0, 0, 0}
	};

	int longIndex = 0;
	int opt = 0;
	std::string type;
	std::string version = "1";

	while ((opt = getopt_long(argc, argv, "t:a:vh", longOptions, &longIndex)) != -1) {
		switch(opt) {
			case 't':
				type.assign(optarg, strlen(optarg));
				break;
			case 'a':
				version.assign(optarg, strlen(optarg));
				break;
			case 'v':
			case 'h':
			default:
				printHelp();
				return 0;
		}
	}

	type = getLowerCaseType(type);

	std::list<sax::Token> tokens;
	sax::SaxParseInterface::parseStdin(tokens);

	try {
		if (type == TYPE_FSM) {
			std::string xmlMark = tokens.front( ).getData( );
			if(xmlMark == "NFA") {
				automaton::NFA nfa = alib::DataFactory::fromTokens<automaton::NFA>(tokens);
				automaton::DFA dfa = determinize::NFADeterminizer::determinize(nfa);
				alib::DataFactory::toStdout(dfa);
				return 0;
			} else if(xmlMark == "DFA") {
				automaton::DFA automaton = alib::DataFactory::fromTokens<automaton::DFA>(tokens);
				alib::DataFactory::toStdout(automaton);
				return 0;
			} else {
				throw exception::AlibException("Unsupported formalism");
			}

		} else if (type == TYPE_IDPDA) {
			automaton::InputDrivenNPDA npda = alib::DataFactory::fromTokens<automaton::InputDrivenNPDA>(tokens);
			automaton::DPDA dpda = determinize::IDPDADeterminizer::determinize(npda);
			alib::DataFactory::toStdout(dpda);
			return 0;
		} else if (type == TYPE_VPA) {
			automaton::VisiblyPushdownNPDA npda = alib::DataFactory::fromTokens<automaton::VisiblyPushdownNPDA>(tokens);
			automaton::VisiblyPushdownDPDA dpda = determinize::VPADeterminizer::determinize(npda);
			alib::DataFactory::toStdout(dpda);
			return 0;
		} else if (type == TYPE_RHDPDA) {
			automaton::NPDA npda = alib::DataFactory::fromTokens<automaton::NPDA>(tokens);
			automaton::RealTimeHeightDeterministicNPDA rhpda = automaton::PDAToRHPDA::convert(npda);
			automaton::RealTimeHeightDeterministicDPDA dpda = determinize::HDPDADeterminizer::determinize(rhpda);
			automaton::DPDA dpda2 = automaton::RHPDAToPDA::convert(dpda);
			alib::DataFactory::toStdout(dpda2);
			return 0;
		}

		printHelp(); // should not be reached
		return 2;
	} catch (const exception::AlibException& e) {
		alib::DataFactory::toStdout(e);
		return 1;
	} catch (...) {
		return 127;
	}
}
