/*
 * PDAToRHPDA.cpp
 *
 *  Created on: 23. 3. 2014
 *	  Author: Jan Travnicek
 */

#include "PDAToRHPDA.h"

#include <exception/AlibException.h>
#include <automaton/PDA/RealTimeHeightDeterministicDPDA.h>
#include <automaton/PDA/RealTimeHeightDeterministicNPDA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/PDA/NPDA.h>

#include <alphabet/BottomOfTheStackSymbol.h>

#include <set>
#include <map>
#include <queue>

namespace automaton {

automaton::RealTimeHeightDeterministicDPDA PDAToRHPDA::convert( const automaton::RealTimeHeightDeterministicDPDA & pda ) {
	return pda;
}

automaton::RealTimeHeightDeterministicNPDA PDAToRHPDA::convert( const automaton::RealTimeHeightDeterministicNPDA & pda ) {
	return pda;
}

automaton::RealTimeHeightDeterministicDPDA PDAToRHPDA::convert( const automaton::DPDA & pda ) {
	automaton::State q0 = automaton::createUniqueState(automaton::State("q0"), pda.getStates());

	RealTimeHeightDeterministicDPDA res(q0, alphabet::Symbol { alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK } );

	res.setInputSymbols(pda.getInputAlphabet());
	res.setStates(pda.getStates());
	res.setFinalStates(pda.getFinalStates());
	std::set<alphabet::Symbol> stackSymbols = pda.getStackAlphabet();
	stackSymbols.insert( alphabet::Symbol { alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK } );
	res.setStackSymbols(stackSymbols);

	res.addCallTransition(q0, pda.getInitialState(), pda.getInitialSymbol());

	std::string us("us");
	int i = 0;
	for(const auto& transition : pda.getTransitions()) {
		const auto& to = transition.second;
		if(std::get<2>(transition.first).size() == 0 && to.second.size() == 0)
			res.addLocalTransition(std::get<0>(transition.first), std::get<1>(transition.first), to.first);
		else if(std::get<2>(transition.first).size() == 1 && to.second.size() == 0)
			res.addReturnTransition(std::get<0>(transition.first), std::get<1>(transition.first), std::get<2>(transition.first)[0], to.first);
		else if(std::get<2>(transition.first).size() == 0 && to.second.size() == 1)
			res.addCallTransition(std::get<0>(transition.first), std::get<1>(transition.first), to.first, to.second[0]);
		else {
			int popPushIndex = 0;
			int popPushSymbols = std::get<2>(transition.first).size() + to.second.size();

			automaton::State lastUS = automaton::createUniqueState(automaton::State(us + std::to_string(i)), res.getStates());
			for(const alphabet::Symbol& pop : std::get<2>(transition.first)) {
				automaton::State fromState = (popPushIndex == 0) ? std::get<0>(transition.first) : lastUS;
				if(popPushIndex != 0) lastUS = automaton::createUniqueState(automaton::State(us + std::to_string(++i)), res.getStates());
				automaton::State toState = (popPushIndex == popPushSymbols - 1) ? to.first : lastUS;

				res.addState(fromState);
				res.addState(toState);

				if(popPushIndex == 0)
					res.addReturnTransition(fromState, std::get<1>(transition.first), pop, toState);
				else
					res.addReturnTransition(fromState, pop, toState);
				popPushIndex++;
			}
			for(const alphabet::Symbol& push : to.second) {
				automaton::State fromState = (popPushIndex == 0) ? std::get<0>(transition.first) : lastUS;
				if(popPushIndex != 0) lastUS = automaton::createUniqueState(automaton::State(us + std::to_string(++i)), res.getStates());
				automaton::State toState = (popPushIndex == popPushSymbols - 1) ? to.first : lastUS;

				res.addState(fromState);
				res.addState(toState);

				if(popPushIndex == 0)
					res.addCallTransition(fromState, std::get<1>(transition.first), toState, push);
				else
					res.addCallTransition(fromState, toState, push);
				popPushIndex++;
			}
		}
	}
	
	return res;
}

automaton::RealTimeHeightDeterministicNPDA PDAToRHPDA::convert( const automaton::NPDA & pda ) {
	RealTimeHeightDeterministicNPDA res(alphabet::Symbol { alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK } );

	res.setInputSymbols(pda.getInputAlphabet());
	res.setStates(pda.getStates());
	res.setFinalStates(pda.getFinalStates());
	std::set<alphabet::Symbol> stackSymbols = pda.getStackAlphabet();
	stackSymbols.insert( alphabet::Symbol { alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK } );
	res.setStackSymbols(stackSymbols);

	automaton::State q0 = automaton::createUniqueState(automaton::State("q0"), res.getStates());
	res.addState(q0);
	res.addInitialState(q0);

	for(const automaton::State& state : pda.getInitialStates())
		for(const alphabet::Symbol& symbol : pda.getInitialSymbols())
			res.addCallTransition(q0, state, symbol);

	std::string us("us");
	int i = 0;
	for(const auto& transition : pda.getTransitions()) {
		for(const auto& to : transition.second) {
			if(std::get<2>(transition.first).size() == 0 && to.second.size() == 0)
				res.addLocalTransition(std::get<0>(transition.first), std::get<1>(transition.first), to.first);
			else if(std::get<2>(transition.first).size() == 1 && to.second.size() == 0)
				res.addReturnTransition(std::get<0>(transition.first), std::get<1>(transition.first), std::get<2>(transition.first)[0], to.first);
			else if(std::get<2>(transition.first).size() == 0 && to.second.size() == 1)
				res.addCallTransition(std::get<0>(transition.first), std::get<1>(transition.first), to.first, to.second[0]);
			else {
				int popPushIndex = 0;
				int popPushSymbols = std::get<2>(transition.first).size() + to.second.size();

				automaton::State lastUS = automaton::createUniqueState(automaton::State(us + std::to_string(i)), res.getStates());
				for(const alphabet::Symbol& pop : std::get<2>(transition.first)) {
					automaton::State fromState = (popPushIndex == 0) ? std::get<0>(transition.first) : lastUS;
					if(popPushIndex != 0) lastUS = automaton::createUniqueState(automaton::State(us + std::to_string(++i)), res.getStates());
					automaton::State toState = (popPushIndex == popPushSymbols - 1) ? to.first : lastUS;

					res.addState(fromState);
					res.addState(toState);

					if(popPushIndex == 0)
						res.addReturnTransition(fromState, std::get<1>(transition.first), pop, toState);
					else
						res.addReturnTransition(fromState, pop, toState);
					popPushIndex++;
				}
				for(const alphabet::Symbol& push : to.second) {
					automaton::State fromState = (popPushIndex == 0) ? std::get<0>(transition.first) : lastUS;
					if(popPushIndex != 0) lastUS = automaton::createUniqueState(automaton::State(us + std::to_string(++i)), res.getStates());
					automaton::State toState = (popPushIndex == popPushSymbols - 1) ? to.first : lastUS;

					res.addState(fromState);
					res.addState(toState);

					if(popPushIndex == 0)
						res.addCallTransition(fromState, std::get<1>(transition.first), toState, push);
					else
						res.addCallTransition(fromState, toState, push);
					popPushIndex++;
				}
			}
		}
	}
	
	return res;
}

automaton::Automaton PDAToRHPDA::convert(const Automaton& automaton) {
	automaton::Automaton* out = NULL;
	automaton.getData().Accept((void*) &out, PDAToRHPDA::PDA_TO_RHPDA);
	automaton::Automaton res(std::move(*out));
	delete out;
	return std::move(res);
}

void PDAToRHPDA::Visit(void*, const UnknownAutomaton&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void PDAToRHPDA::Visit(void*, const EpsilonNFA&) const {
	throw exception::AlibException("Unsupported automaton type EpsilonNFA");
}

void PDAToRHPDA::Visit(void*, const NFA&) const {
	throw exception::AlibException("Unsupported automaton type NFA");
}

void PDAToRHPDA::Visit(void*, const DFA&) const {
	throw exception::AlibException("Unsupported automaton type DFA");
}

void PDAToRHPDA::Visit(void*, const ExtendedNFA&) const {
	throw exception::AlibException("Unsupported automaton type ExtendedDFA");
}

void PDAToRHPDA::Visit(void*, const CompactNFA&) const {
	throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void PDAToRHPDA::Visit(void* data, const DPDA& automaton) const {
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

void PDAToRHPDA::Visit(void*, const SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void PDAToRHPDA::Visit(void*, const InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void PDAToRHPDA::Visit(void*, const VisiblyPushdownDPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void PDAToRHPDA::Visit(void*, const VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void PDAToRHPDA::Visit(void* data, const RealTimeHeightDeterministicDPDA& automaton) const {
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

void PDAToRHPDA::Visit(void* data, const RealTimeHeightDeterministicNPDA& automaton) const {
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

void PDAToRHPDA::Visit(void* data, const NPDA& automaton) const {
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

void PDAToRHPDA::Visit(void*, const SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void PDAToRHPDA::Visit(void*, const OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

const PDAToRHPDA PDAToRHPDA::PDA_TO_RHPDA;

}

