/*
 * StateEliminationFormal.cpp
 *
 *  Created on: 9. 2. 2014
 *      Author: Tomas Pecka
 */

#include "StateEliminationFormal.h"

#include <regexp/formal/FormalRegExp.h>
#include <exception/AlibException.h>

#include "../../../regexp/RegExpOptimize.h"

namespace conversions
{

namespace fa2re
{

regexp::RegExp StateEliminationFormal::convert(const automaton::Automaton& automaton)
{
    regexp::RegExp* out = NULL;
    automaton.getData().Accept((void*) &out, StateEliminationFormal::STATE_ELIMINATION_FORMAL);
    regexp::RegExp res = std::move(*out);
    delete out;
    return res;
}

template<class T>
regexp::FormalRegExp StateEliminationFormal::convert(const T& automaton)
{
    if(automaton.getFinalStates().size() == 0)
        return regexp::FormalRegExp(regexp::FormalRegExpEmpty());

    // steps 1 + 2
    automaton::ExtendedNFA extendedAutomaton = constructExtendedNFA(automaton);
    extendExtendedNFA(extendedAutomaton);

    // step 3 - Exterminate!
    // select all states that are neither final nor initial
    std::set<automaton::State> statesToEliminate = extendedAutomaton.getStates();
    statesToEliminate.erase(*extendedAutomaton.getInitialStates().begin());
    statesToEliminate.erase(*extendedAutomaton.getFinalStates().begin());

    for(const auto& state : statesToEliminate)
        extendedAutomaton = eliminateState(extendedAutomaton, state);

    // step 4
    //regexp::RegExpOptimize opt;

    regexp::FormalRegExpConcatenation concat(
        std::move(*transition(extendedAutomaton, *extendedAutomaton.getInitialStates().begin(), *extendedAutomaton.getFinalStates().begin())),
        regexp::FormalRegExpIteration(std::move(*transition(extendedAutomaton, *extendedAutomaton.getFinalStates().begin(), *extendedAutomaton.getFinalStates().begin())))
        );
    //return regexp::RegExp{opt.optimize(regexp::FormalRegExp(concat))};
    return regexp::FormalRegExp(concat);
}


void StateEliminationFormal::Visit(void*, const automaton::UnknownAutomaton&) const
{
    throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void StateEliminationFormal::Visit(void* data, const automaton::EpsilonNFA& automaton) const
{
    regexp::RegExp* & out = *((regexp::RegExp**) data);
    out = new regexp::RegExp(convert(automaton));
}

void StateEliminationFormal::Visit(void* data, const automaton::NFA& automaton) const
{
    regexp::RegExp* & out = *((regexp::RegExp**) data);
    out = new regexp::RegExp(convert(automaton));
}

void StateEliminationFormal::Visit(void* data, const automaton::DFA& automaton) const
{
    regexp::RegExp* & out = *((regexp::RegExp**) data);
    out = new regexp::RegExp(convert(automaton));
}

void StateEliminationFormal::Visit(void* data, const automaton::ExtendedNFA& automaton) const
{
    regexp::RegExp* & out = *((regexp::RegExp**) data);
    out = new regexp::RegExp(convert(automaton));
}

void StateEliminationFormal::Visit(void*, const automaton::CompactNFA& ) const
{
    throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void StateEliminationFormal::Visit(void*, const automaton::DPDA&) const
{
    throw exception::AlibException("Unsupported automaton type DPDA");
}

void StateEliminationFormal::Visit(void*, const automaton::SinglePopDPDA&) const
{
    throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void StateEliminationFormal::Visit(void*, const automaton::InputDrivenNPDA&) const
{
    throw exception::AlibException("Unsupported automaton type NPDA");
}

void StateEliminationFormal::Visit(void*, const automaton::VisiblyPushdownDPDA&) const
{
    throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void StateEliminationFormal::Visit(void*, const automaton::VisiblyPushdownNPDA&) const
{
    throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void StateEliminationFormal::Visit(void*, const automaton::RealTimeHeightDeterministicDPDA&) const
{
    throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void StateEliminationFormal::Visit(void*, const automaton::RealTimeHeightDeterministicNPDA&) const
{
    throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void StateEliminationFormal::Visit(void*, const automaton::NPDA&) const
{
    throw exception::AlibException("Unsupported automaton type NPDA");
}

void StateEliminationFormal::Visit(void*, const automaton::SinglePopNPDA&) const
{
    throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void StateEliminationFormal::Visit(void*, const automaton::OneTapeDTM&) const
{
    throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}


automaton::ExtendedNFA StateEliminationFormal::eliminateState(const automaton::ExtendedNFA& extendedAutomaton, const automaton::State& q)
{
    automaton::ExtendedNFA newAutomaton;
    newAutomaton.setStates(extendedAutomaton.getStates());
    newAutomaton.removeState(q); // preserve all states but q (the one to eliminate)
    newAutomaton.setInputSymbols(extendedAutomaton.getInputAlphabet());
    newAutomaton.setInitialStates(extendedAutomaton.getInitialStates()); // sure that q is neither initial nor final (follows from step 2 - extending ExtendedNFA)
    newAutomaton.setFinalStates(extendedAutomaton.getFinalStates());

    // regexp::RegExpOptimize opt;

    for(const auto& p: newAutomaton.getStates())
    {
        for(const auto& r : newAutomaton.getStates())
        {
            regexp::FormalRegExpConcatenation concat(
                std::move(*transition(extendedAutomaton, p, q)),
                regexp::FormalRegExpConcatenation(
                    regexp::FormalRegExpIteration(std::move(*transition(extendedAutomaton, q, q))),
                    std::move(*transition(extendedAutomaton, q, r))
                )
            );

            regexp::FormalRegExpAlternation alt(
                std::move(*transition(extendedAutomaton, p, r)),
                concat
            );

            // regexp::FormalRegExp transitionRegExp(opt.optimize(regexp::FormalRegExp(alt)));
            newAutomaton.addTransition(p, regexp::RegExp{regexp::FormalRegExp(alt)}, r);
        }
    }

    return newAutomaton;
}

const regexp::FormalRegExpElement* StateEliminationFormal::transition(const automaton::ExtendedNFA& automaton, const automaton::State& from, const automaton::State& to)
{
    std::vector<regexp::FormalRegExpElement*> alt;

    for(const auto& transition: automaton.getTransitionsFromState(from))
        if(transition.second.count(to) > 0)
            alt.push_back(static_cast<const regexp::FormalRegExp&>(transition.first.second.getData()).getRegExp().clone());

    size_t size = alt.size();
    if(size == 0) return new regexp::FormalRegExpEmpty();
    if(size == 1) return alt.at(0);

    regexp::FormalRegExpAlternation* ret = new regexp::FormalRegExpAlternation(std::move(*alt.at(--size)), std::move(*alt.at(--size)));
    while(size)
    {
        regexp::FormalRegExpElement* left = alt.at(--size);
        regexp::FormalRegExpElement* right = ret;

        ret = new regexp::FormalRegExpAlternation(std::move(*left), std::move(*right));
        delete left;
        delete right;
    }

    return ret;
}

template<>
automaton::ExtendedNFA StateEliminationFormal::constructExtendedNFA(const automaton::NFA& automaton)
{
    automaton::ExtendedNFA extendedAutomaton;
    extendedAutomaton.setStates(automaton.getStates());
    extendedAutomaton.setInputSymbols(automaton.getInputAlphabet());
    extendedAutomaton.setInitialStates(automaton.getInitialStates());
    extendedAutomaton.setFinalStates(automaton.getFinalStates());

    for(const auto& transition : automaton.getTransitions()) // pair<State, symb/eps> -> set<State>
        for(const auto& to : transition.second)
            extendedAutomaton.addTransition(transition.first.first, regexp::RegExp{regexp::FormalRegExp(regexp::FormalRegExpSymbol(transition.first.second))}, to);

    return extendedAutomaton;
}

template<>
automaton::ExtendedNFA StateEliminationFormal::constructExtendedNFA(const automaton::DFA& automaton)
{
    automaton::ExtendedNFA extendedAutomaton;
    extendedAutomaton.setStates(automaton.getStates());
    extendedAutomaton.setInputSymbols(automaton.getInputAlphabet());
    extendedAutomaton.setInitialStates(std::set<automaton::State>{automaton.getInitialState()});
    extendedAutomaton.setFinalStates(automaton.getFinalStates());

    for(const auto& transition : automaton.getTransitions()) // pair<State, symb/eps> -> set<State>
        extendedAutomaton.addTransition(transition.first.first, regexp::RegExp{regexp::FormalRegExp(regexp::FormalRegExpSymbol(transition.first.second))}, transition.second);

    return extendedAutomaton;
}

template<>
automaton::ExtendedNFA StateEliminationFormal::constructExtendedNFA(const automaton::EpsilonNFA& automaton)
{
    automaton::ExtendedNFA extendedAutomaton;
    extendedAutomaton.setStates(automaton.getStates());
    extendedAutomaton.setInputSymbols(automaton.getInputAlphabet());
    extendedAutomaton.setInitialStates(automaton.getInitialStates());
    extendedAutomaton.setFinalStates(automaton.getFinalStates());

    for(const auto& transition : automaton.getSymbolTransitions()) // pair<State, symb/eps> -> set<State>
        for(const auto& to : transition.second)
            extendedAutomaton.addTransition(transition.first.first, regexp::RegExp{regexp::FormalRegExp(regexp::FormalRegExpSymbol(transition.first.second))}, to);

    for(const auto& transition : automaton.getEpsilonTransitions())
        for(const auto& to : transition.second)
            extendedAutomaton.addTransition(transition.first, regexp::RegExp{regexp::FormalRegExp(regexp::FormalRegExpEpsilon())}, to);

    return extendedAutomaton;
}

template<>
automaton::ExtendedNFA StateEliminationFormal::constructExtendedNFA(const automaton::ExtendedNFA& automaton)
{
    return automaton;
}

void StateEliminationFormal::extendExtendedNFA(automaton::ExtendedNFA& automaton)
{
    const automaton::State& initState = *automaton.getInitialStates().begin();
    if(automaton.getFinalStates().count(initState) > 0 || automaton.getTransitionsToState(initState).size() > 0 )
    {
        const automaton::State q0 = automaton::createUniqueState(initState, automaton.getStates());
        automaton.addState(q0);

        regexp::FormalRegExp regexp;
        regexp.setRegExp(regexp::FormalRegExpEpsilon());
        automaton.addTransition(q0, regexp::RegExp{regexp}, initState);

        const std::set<automaton::State> automatonInitialStates = automaton.getInitialStates();
        for(const auto& state: automatonInitialStates)
            automaton.removeInitialState(state);
        automaton.addInitialState(q0);
    }

    if(automaton.getFinalStates().size() > 1)
    {
        const automaton::State f = automaton::createUniqueState(automaton::State("f"), automaton.getStates());
        automaton.addState(f);

        const std::set<automaton::State> automatonFinalStates = automaton.getFinalStates();
        for(const auto &state : automatonFinalStates)
        {
            regexp::FormalRegExp regexp;
            regexp.setRegExp(regexp::FormalRegExpEpsilon());
            automaton.addTransition(state, regexp::RegExp{regexp}, f);

            automaton.removeFinalState(state);
        }

        automaton.addFinalState(f);
    }
}

const StateEliminationFormal StateEliminationFormal::STATE_ELIMINATION_FORMAL;

} /* namespace fa2re */

} /* namespace conversions */
