/*
 * BrzozowskiDerivation.cpp
 *
 *  Created on: 11. 1. 2014
 *	  Author: Tomas Pecka
 */

#include "BrzozowskiDerivation.h"

#include <set>
#include <deque>
#include <queue>
#include <vector>

#include <string/LinearString.h>
#include <std/hexavigesimal.h>

#include "../../regexp/RegExpDerivation.h"
#include "../../regexp/RegExpOptimize.h"
#include "../../regexp/RegExpEpsilon.h"

namespace conversions
{

namespace re2fa
{

automaton::Automaton BrzozowskiDerivation::convert(const regexp::RegExp& regexp)
{
	automaton::Automaton* out = NULL;
	regexp.getData().Accept((void*) &out, BrzozowskiDerivation::BRZOZOWSKI_DERIVATION);
	automaton::Automaton res = std::move(*out);
	delete out;
	return res;
}

template<class T>
automaton::NFA BrzozowskiDerivation::convert(const T& regexp)
{
	// 1.
	regexp::RegExpOptimize opt;
	regexp::RegExp V = regexp::RegExp{opt.optimize(regexp)};

	std::set<regexp::RegExp> Q = { V };
	std::deque<std::set<regexp::RegExp>> Qi;

	Qi.push_back(std::set<regexp::RegExp>());
	Qi.at(0).insert(V);

	int i = 1;

	// 2.
	while(! Qi.at(i - 1).empty())
	{
		Qi.push_back(std::set<regexp::RegExp>()); // initialize set Q_i

		for(const auto& dregexp : Qi.at(i - 1))
		{
			regexp::RegExpDerivation deriv;

			for(const auto& a : regexp.getAlphabet())
			{
				string::LinearString string(std::vector<alphabet::Symbol>{a});
				regexp::RegExp derived = deriv.derivation(dregexp, string);
				derived = opt.optimize(derived);

				// this will also add \emptyset as a regexp (and as FA state)
				if(Q.count(derived) == 0) // if this state has already been found, do not add
					Qi.at(i).insert(derived);
			}
		}

		Q.insert(Qi.at(i).begin(), Qi.at(i).end());
		i += 1;
	}

	std::map<regexp::RegExp, automaton::State> stateMap;
	int stateId = 0;

	for(const auto& r : Q)
	{
		automaton::State q(std::toBase26(stateId++));
		stateMap.insert(std::make_pair(r, q));
	}

	// ------------------------------------------------------------------------
	// 3.

	automaton::NFA automaton;

	for(const auto& r : stateMap)
	{
		automaton.addState(r.second);
	}

	automaton.addInitialState(stateMap.find(V)->second);

	automaton.setInputSymbols(regexp.getAlphabet());

	for(const auto& r : Q)
	{
		regexp::RegExpDerivation deriv;

		for(const auto& a: regexp.getAlphabet())
		{
			string::LinearString string(std::vector<alphabet::Symbol>{a});
			regexp::RegExp derived = deriv.derivation(r, string);
			derived = opt.optimize(derived);

			automaton.addTransition(stateMap.find(r)->second, a, stateMap.find(derived)->second);
		}
	}

	for(const auto& r : Q)
	{
		if(regexp::RegExpEpsilon::languageContainsEpsilon(r))
			automaton.addFinalState(stateMap.find(r)->second);
	}

	return automaton;
}

template automaton::NFA BrzozowskiDerivation::convert(const regexp::FormalRegExp& regexp);
template automaton::NFA BrzozowskiDerivation::convert(const regexp::UnboundedRegExp& regexp);

void BrzozowskiDerivation::Visit(void* userData, const regexp::FormalRegExp& regexp) const
{
	automaton::Automaton* &out = *((automaton::Automaton**) userData);
	out = new automaton::Automaton(this->convert(regexp));
}
void BrzozowskiDerivation::Visit(void* userData, const regexp::UnboundedRegExp& regexp) const
{
	automaton::Automaton* &out = *((automaton::Automaton**) userData);
	out = new automaton::Automaton(this->convert(regexp));
}

const BrzozowskiDerivation BrzozowskiDerivation::BRZOZOWSKI_DERIVATION;

} /* namespace re2fa */

} /* namespace conversions */
