/*
 * NFADeterminizer.cpp
 *
 *  Created on: 16. 1. 2014
 *	  Author: Jan Vesely
 */

#include "NFADeterminizer.h"
#include "../common/NFACommon.h"

#include <deque>
#include <algorithm>

namespace determinize {

automaton::Automaton NFADeterminizer::determinize(const automaton::Automaton& automaton) {
	automaton::Automaton* out = NULL;
	automaton.getData().Accept((void*) &out, NFADeterminizer::NFA_DETERMINIZER);
	automaton::Automaton res = std::move(*out);
	delete out;
	return res;
}

automaton::DFA NFADeterminizer::determinize(const automaton::NFA& nfa) {
	// 1, 4
	automaton::State initialState(createDFAState(nfa.getInitialStates()));
	automaton::DFA res(initialState);
	res.setInputSymbols(nfa.getInputAlphabet());
	
	// 2
	std::deque<automaton::State> todo;
	todo.push_back(initialState);
	
	do {
		// 3a, c
		automaton::State state = todo.front();
		todo.pop_front();

		// 3b
		for (const auto& input : nfa.getInputAlphabet()) {
			std::set<automaton::State> targetNFAStates;
			for(const auto& nfaState : recreateNFAStates(state)) {
				auto iter = nfa.getTransitions().find(std::make_pair(nfaState, input));
				if(iter != nfa.getTransitions().end()) {
					targetNFAStates.insert(iter->second.begin(), iter->second.end());
				}
			}
			automaton::State dfaState = createDFAState(targetNFAStates);

			// 4
			bool existed = !res.addState(dfaState);

			// 3b
			res.addTransition(state, input, dfaState);

			if(!existed) todo.push_back(dfaState);
		}
	} while(!todo.empty());
	
	// 5
	for (const auto& dfaState : res.getStates()) {
		std::set<automaton::State> nfaStates = recreateNFAStates(dfaState);
		if(std::any_of(nfaStates.begin(), nfaStates.end(), [&](const automaton::State& nfaState) { return nfa.getFinalStates().count(nfaState); })) {
			res.addFinalState(dfaState);
		}
	}
	
	return res;
}

void NFADeterminizer::Visit(void*, const automaton::UnknownAutomaton&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void NFADeterminizer::Visit(void*, const automaton::EpsilonNFA&) const {
	throw exception::AlibException("Unsupported automaton type EpsilonNFA");
}

void NFADeterminizer::Visit(void* data, const automaton::NFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->determinize(automaton));
}

void NFADeterminizer::Visit(void*, const automaton::DFA&) const {
	throw exception::AlibException("Unsupported automaton type DFA");
}

void NFADeterminizer::Visit(void*, const automaton::ExtendedNFA& ) const {
	throw exception::AlibException("Unsupported automaton type ExtendedNFA");
}

void NFADeterminizer::Visit(void*, const automaton::CompactNFA& ) const {
	throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void NFADeterminizer::Visit(void*, const automaton::DPDA&) const {
	throw exception::AlibException("Unsupported automaton type DPDA");
}

void NFADeterminizer::Visit(void*, const automaton::SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void NFADeterminizer::Visit(void*, const automaton::InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void NFADeterminizer::Visit(void*, const automaton::VisiblyPushdownDPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void NFADeterminizer::Visit(void*, const automaton::VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void NFADeterminizer::Visit(void*, const automaton::RealTimeHeightDeterministicDPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void NFADeterminizer::Visit(void*, const automaton::RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void NFADeterminizer::Visit(void*, const automaton::NPDA&) const {
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void NFADeterminizer::Visit(void*, const automaton::SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void NFADeterminizer::Visit(void*, const automaton::OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

const NFADeterminizer NFADeterminizer::NFA_DETERMINIZER;

}
