/*
 * RandomRegExpFactory.cpp
 *
 *  Created on: 27. 3. 2014
 *	  Author: Jan Travnicek
 */

#include "RandomRegExpFactory.h"

#include <algorithm>

namespace generator {

regexp::UnboundedRegExp RandomRegExpFactory::generateUnboundedRegExp( size_t leafNodes, size_t height, size_t alphabetSize ) {
	srand( time( NULL ) );

	if(alphabetSize > 26)
		throw exception::AlibException("Too big alphabet.");

	if( alphabetSize <= 0 )
		throw exception::AlibException( "Alphabet size must be greater than 0." );

	std::set<alphabet::Symbol> alphabet;
	while( alphabet.size( ) < alphabetSize ) {
		std::string s( 1, rand() % 26 + 'a' );
		alphabet::Symbol symbol = alphabet::symbolFrom (s);
		alphabet.insert( symbol );
	}

	return RandomRegExpFactory::generateUnboundedRegExp( leafNodes, height, alphabet );
}

regexp::UnboundedRegExp RandomRegExpFactory::generateUnboundedRegExp( size_t leafNodes, size_t height, std::set<alphabet::Symbol> alphabet) {

	if( alphabet.size() > 26)
		throw exception::AlibException("Too big alphabet.");

	if( alphabet.size() <= 0 )
		throw exception::AlibException( "Alphabet size must be greater than 0." );

	std::vector<const regexp::UnboundedRegExpElement*> elems;

	{
		elems.push_back(new regexp::UnboundedRegExpEmpty());
		elems.push_back(new regexp::UnboundedRegExpEpsilon());
	}
	if(alphabet.size() > 6) {
		elems.push_back(new regexp::UnboundedRegExpEmpty());
		elems.push_back(new regexp::UnboundedRegExpEpsilon());
	}
	if(alphabet.size() > 16) {
		elems.push_back(new regexp::UnboundedRegExpEmpty());
		elems.push_back(new regexp::UnboundedRegExpEpsilon());
	}

	for(const alphabet::Symbol& symbol : alphabet) {
		elems.push_back(new regexp::UnboundedRegExpSymbol(symbol));
	}

	regexp::UnboundedRegExp res = RandomRegExpFactory::SimpleUnboundedRegExp( leafNodes, height, elems );

	for(const regexp::UnboundedRegExpElement* elem : elems ) {
		delete elem;
	}
	return res;
}

regexp::UnboundedRegExp RandomRegExpFactory::SimpleUnboundedRegExp( size_t n, size_t h, const std::vector<const regexp::UnboundedRegExpElement*> & elems) {

	const regexp::UnboundedRegExpElement* elem = SimpleUnboundedRegExpElement(n, h, elems);
	regexp::UnboundedRegExp res(std::move(*elem));
	delete elem;
	return res;
}

const regexp::UnboundedRegExpElement* RandomRegExpFactory::SimpleUnboundedRegExpElement(size_t n, size_t h, const std::vector<const regexp::UnboundedRegExpElement*> & elems) {
	if(h == 0 || n == 0) {
		return elems[ rand( ) % elems.size( ) ]->clone();
	} else {
		unsigned childNodes = rand( ) % 10;
		if(childNodes <  3) childNodes = 1;
		else if(childNodes <  6) childNodes = 2;
		else if(childNodes <  8) childNodes = 3;
		else childNodes = 4;

		childNodes = childNodes > n ? n : childNodes;

		int subSizes[4];
		if(childNodes == 4) {
			subSizes[3] = rand() % ( n - 1 );
			subSizes[2] = rand() % ( n - subSizes[3] - 1 );
			subSizes[1] = rand() % ( n - subSizes[2] - subSizes [3] - 1 );

			subSizes[3] += 1;
			subSizes[2] += 1;
			subSizes[1] += 1;

			subSizes[0] = n - subSizes[1] - subSizes[2] - subSizes[3];
		}
		if(childNodes == 3) {
			subSizes[2] = rand() % ( n - 1);
			subSizes[1] = rand() % ( n - subSizes[2] - 1);

			subSizes[2] += 1;
			subSizes[1] += 1;

			subSizes[0] = n - subSizes[1] - subSizes[2];
		}
		if(childNodes == 2) {
			subSizes[1] = rand() % ( n - 1 );

			subSizes[1] += 1;

			subSizes[0] = n - subSizes[1];
		}
		if(childNodes == 1) {
			const regexp::UnboundedRegExpElement* subElem = SimpleUnboundedRegExpElement(n, h - 1, elems);
			regexp::UnboundedRegExpIteration iter(std::move(*subElem));
			delete subElem;
			return new regexp::UnboundedRegExpIteration(iter);
		}

		int nodeType = rand( ) % 2;
		if(nodeType == 0) {
			regexp::UnboundedRegExpConcatenation con;
			for(unsigned i = 0; i < childNodes; i++) {
				const regexp::UnboundedRegExpElement* subElem = SimpleUnboundedRegExpElement(subSizes[i], h - 1, elems);
				con.appendElement(std::move(*subElem));
				delete subElem;
			}
			return new regexp::UnboundedRegExpConcatenation(con);
		} else {
			regexp::UnboundedRegExpAlternation con;
			for(unsigned i = 0; i < childNodes; i++) {
				const regexp::UnboundedRegExpElement* subElem = SimpleUnboundedRegExpElement(subSizes[i], h - 1, elems);
				con.appendElement(std::move(*subElem));
				delete subElem;
			}
			return new regexp::UnboundedRegExpAlternation(con);
		}
		
	}
}

}
