/*
 * MinimizeDFA.cpp
 *
 *  Created on: Dec 9, 2013
 *      Author: honza
 */

#include "MinimizeDFA.h"

#include <map>
#include <set>
#include <sstream>
#include <iostream>

#include <alphabet/Symbol.h>

#include <exception/AlibException.h>

#include <automaton/Automaton.h>

namespace minimize {

automaton::Automaton MinimizeDFA::minimize(const automaton::Automaton& automaton) {
	automaton::Automaton* out = NULL;
	automaton.getData().Accept((void*) &out, MinimizeDFA::MINIMIZE_DFA);
	automaton::Automaton res = std::move(*out);
	delete out;
	return res;
}

automaton::DFA MinimizeDFA::minimize(const automaton::DFA& dfa) {
	if(dfa.getFinalStates().size() == 0) {
		automaton::DFA result(automaton::State(0));
		result.setInputSymbols(dfa.getInputAlphabet());
		return result;
	}

	std::map<automaton::State, std::map<alphabet::Symbol, automaton::State > > refactor;

	for(std::set<automaton::State>::const_iterator iter = dfa.getStates().begin(); iter != dfa.getStates().end(); iter++) {
		refactor.insert(std::pair<automaton::State, std::map<alphabet::Symbol, automaton::State> >(*iter, std::map<alphabet::Symbol, automaton::State>()));
	}
	
	for(std::map<std::pair<automaton::State, alphabet::Symbol>, automaton::State>::const_iterator iter = dfa.getTransitions().begin(); iter != dfa.getTransitions().end(); iter++) {
		refactor[iter->first.first].insert(std::pair<alphabet::Symbol, automaton::State>(iter->first.second, iter->second));
	}

	std::map<automaton::State, automaton::State> toEquvivalentStates1;//original state to equivalent state
	std::map<std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > >, std::set<automaton::State> > minimizedTransitionFunction1; //mapped to the original state

	std::map<automaton::State, automaton::State> toEquvivalentStates2;
	std::map<std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > >, std::set<automaton::State> > minimizedTransitionFunction2;
	
	for(auto iter = dfa.getStates().begin(); iter != dfa.getStates().end(); iter++) {
		if(dfa.getFinalStates().count(*iter) == 0) { // not a final state
			toEquvivalentStates2.insert(std::pair<automaton::State, automaton::State>(*iter, automaton::State(0)));
		} else {
			toEquvivalentStates2.insert(std::pair<automaton::State, automaton::State>(*iter, automaton::State(1)));
		}
	}

	for(auto iter = refactor.begin(); iter != refactor.end(); iter++) {
		automaton::State from = toEquvivalentStates2.find(iter->first)->second;
		std::set<std::pair<alphabet::Symbol, automaton::State> > transitionFunction;

		for(auto iter2 = iter->second.begin(); iter2 != iter->second.end(); iter2++) {
			transitionFunction.insert(std::pair<alphabet::Symbol, automaton::State>(iter2->first, toEquvivalentStates2.find(iter2->second)->second));
		}
		
		std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > > key(from, transitionFunction);
		
		if(minimizedTransitionFunction2.find(key) == minimizedTransitionFunction2.end()) {
			minimizedTransitionFunction2.insert(std::pair<std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > >, std::set<automaton::State> >(key, std::set<automaton::State>()));
		}
		minimizedTransitionFunction2[key].insert(iter->first);
	}
	
	do {
		toEquvivalentStates1 = toEquvivalentStates2;
		minimizedTransitionFunction1 = minimizedTransitionFunction2;

		toEquvivalentStates2.clear();
		minimizedTransitionFunction2.clear();
		
		int number = 0;
		for(auto iter = minimizedTransitionFunction1.begin(); iter != minimizedTransitionFunction1.end(); iter++) {
			for(auto iter2 = iter->second.begin(); iter2 != iter->second.end(); iter2++) {
				toEquvivalentStates2.insert(std::pair<automaton::State, automaton::State>(*iter2, automaton::State(number)));
			}
			number++;
		}

		for(auto iter = refactor.begin(); iter != refactor.end(); iter++) {
			automaton::State from = toEquvivalentStates2.find(iter->first)->second;
			std::set<std::pair<alphabet::Symbol, automaton::State> > transitionFunction;
			
			for(auto iter2 = iter->second.begin(); iter2 != iter->second.end(); iter2++) {
				transitionFunction.insert(std::pair<alphabet::Symbol, automaton::State>(iter2->first, toEquvivalentStates2.find(iter2->second)->second));
			}
			
			std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > > key(from, transitionFunction);
			
			if(minimizedTransitionFunction2.find(key) == minimizedTransitionFunction2.end()) {
				minimizedTransitionFunction2.insert(std::pair<std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > >, std::set<automaton::State> >(key, std::set<automaton::State>()));
			}

			minimizedTransitionFunction2[key].insert(iter->first);
			
		}

	} while(minimizedTransitionFunction1.size() != minimizedTransitionFunction2.size());

	const automaton::State* initialState = NULL;
	for(auto iter = minimizedTransitionFunction2.begin(); iter != minimizedTransitionFunction2.end(); iter++) {
		if(iter->second.find(dfa.getInitialState()) != iter->second.end()) {
			initialState = &(iter->first.first);
			break;
		}
	}

	automaton::DFA result(*initialState);

	for(auto iter = dfa.getInputAlphabet().begin(); iter != dfa.getInputAlphabet().end(); iter++) {
		result.addInputSymbol(*iter);
	}
	
	for(auto iter = minimizedTransitionFunction2.begin(); iter != minimizedTransitionFunction2.end(); iter++) {
		result.addState(iter->first.first);
		if(dfa.getFinalStates().find(*(iter->second.begin())) != dfa.getFinalStates().end()) {
			result.addFinalState(iter->first.first);
		}
	}
	
	for(auto iter = minimizedTransitionFunction2.begin(); iter != minimizedTransitionFunction2.end(); iter++) {
		for(auto iter2 = iter->first.second.begin(); iter2 != iter->first.second.end(); iter2++) {
			result.addTransition(iter->first.first, iter2->first, iter2->second);
		}
	}
	
	return result;
}

void MinimizeDFA::Visit(void*, const automaton::UnknownAutomaton&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void MinimizeDFA::Visit(void*, const automaton::EpsilonNFA&) const {
	throw exception::AlibException("Unsupported automaton type EpsilonNFA");
}

void MinimizeDFA::Visit(void* data, const automaton::DFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->minimize(automaton));
}

void MinimizeDFA::Visit(void*, const automaton::NFA&) const {
	throw exception::AlibException("Unsupported automaton type DFA");
}

void MinimizeDFA::Visit(void*, const automaton::ExtendedNFA& ) const {
	throw exception::AlibException("Unsupported automaton type ExtendedNFA");
}

void MinimizeDFA::Visit(void*, const automaton::CompactNFA& ) const {
	throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void MinimizeDFA::Visit(void*, const automaton::DPDA&) const {
	throw exception::AlibException("Unsupported automaton type DPDA");
}

void MinimizeDFA::Visit(void*, const automaton::SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void MinimizeDFA::Visit(void*, const automaton::InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void MinimizeDFA::Visit(void*, const automaton::VisiblyPushdownDPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void MinimizeDFA::Visit(void*, const automaton::VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void MinimizeDFA::Visit(void*, const automaton::RealTimeHeightDeterministicDPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void MinimizeDFA::Visit(void*, const automaton::RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void MinimizeDFA::Visit(void*, const automaton::NPDA&) const {
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void MinimizeDFA::Visit(void*, const automaton::SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void MinimizeDFA::Visit(void*, const automaton::OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

const MinimizeDFA MinimizeDFA::MINIMIZE_DFA;

} /* namespace minimize */
