/*
 * TrimCFG.cpp
 *
 *  Created on: 22. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "TrimCFG.h"

#include <grammar/ContextFree/CFG.h>
#include <grammar/ContextFree/EpsilonFreeCFG.h>
#include <grammar/ContextFree/GNF.h>
#include <grammar/ContextFree/CNF.h>
#include <grammar/ContextFree/LG.h>
#include <grammar/Regular/LeftLG.h>
#include <grammar/Regular/LeftRG.h>
#include <grammar/Regular/RightLG.h>
#include <grammar/Regular/RightRG.h>

#include <std/set.hpp>

#include "../../grammar/GrammarPropertiesCFG.h"

namespace trim {

template<class T>
T TrimCFG::removeUnreachableSymbols( const T & grammar) {
	// 1.
	std::set<alphabet::Symbol> Vt = grammar::GrammarPropertiesCFG::getUnreachableSymbols( grammar );

	T ret(grammar.getInitialSymbol( ) );

	std::set<alphabet::Symbol> newNonTerminals, newTerminals;

	set_intersection( Vt.begin( ), Vt.end( ), grammar.getNonterminalAlphabet( ).begin( ), grammar.getNonterminalAlphabet( ).end( ), std::inserter( newNonTerminals, newNonTerminals.begin( ) ) );
	for( const auto & symbol : newNonTerminals )
		ret.addNonterminalSymbol( symbol );

	set_intersection( Vt.begin( ), Vt.end( ), grammar.getTerminalAlphabet( ).begin( ), grammar.getTerminalAlphabet( ).end( ), std::inserter( newTerminals, newTerminals.begin( ) ) );
	for( const auto & symbol : newTerminals )
		ret.addTerminalSymbol( symbol );

	// A->\alpha: if A \in N' and \alpha in V_i*, then A->\alpha in P
	for( const auto & rule : grammar.getRawRules( ) ) {
		if( newNonTerminals.count( rule.first ) ) {
			for( const auto& rhs : rule.second ) {
				if( all_of( rhs.begin( ), rhs.end( ), [ Vt ]( alphabet::Symbol const& symb ) -> bool {
					return Vt.count( symb );
				} ) )
					ret.addRawRule( rule.first, rhs );
			}
	}	
	}

	// 2.
	return ret;
}

template grammar::CFG TrimCFG::removeUnreachableSymbols( const grammar::CFG & grammar );
template grammar::EpsilonFreeCFG TrimCFG::removeUnreachableSymbols( const grammar::EpsilonFreeCFG & grammar );
template grammar::GNF TrimCFG::removeUnreachableSymbols( const grammar::GNF & grammar );
template grammar::CNF TrimCFG::removeUnreachableSymbols( const grammar::CNF & grammar );
template grammar::LG TrimCFG::removeUnreachableSymbols( const grammar::LG & grammar );
template grammar::LeftLG TrimCFG::removeUnreachableSymbols( const grammar::LeftLG & grammar );
template grammar::LeftRG TrimCFG::removeUnreachableSymbols( const grammar::LeftRG & grammar );
template grammar::RightLG TrimCFG::removeUnreachableSymbols( const grammar::RightLG & grammar );
template grammar::RightRG TrimCFG::removeUnreachableSymbols( const grammar::RightRG & grammar );

template<class T>
T TrimCFG::removeUnproductiveSymbols( const T & grammar ) {
	// 1.
	std::set<alphabet::Symbol> Nt = grammar::GrammarPropertiesCFG::getProductiveNonterminals( grammar );

	T ret(grammar.getInitialSymbol( ) );

	for( const auto & symbol : Nt )
		ret.addNonterminalSymbol( symbol );

	for( const auto & symbol : grammar.getTerminalAlphabet( ) )
		ret.addTerminalSymbol( symbol );

	const std::set<alphabet::Symbol> & terminals = ret.getTerminalAlphabet( );
	for( const auto & rule : grammar.getRawRules( ) ) {
		if( Nt.count( rule.first ) ) {
			for( const auto & rhs : rule.second ) {
				if( all_of( rhs.begin( ), rhs.end( ), [ Nt, terminals ]( const alphabet::Symbol & symbol ) {
						return Nt.count( symbol ) || terminals.count( symbol );
				} ) )
					ret.addRawRule( rule.first, rhs );
			}
		}
	}


	/* if( ! G1.getNonTerminalSymbols( ) . count( grammar.getInitialSymbol( ) ) )
		throw AlibException( "Starting symbol of grammar was marked as unproductive and therefore it was removed." ); */

	// 2.
	return ret;
}

template grammar::CFG TrimCFG::removeUnproductiveSymbols( const grammar::CFG & grammar );
template grammar::EpsilonFreeCFG TrimCFG::removeUnproductiveSymbols( const grammar::EpsilonFreeCFG & grammar );
template grammar::GNF TrimCFG::removeUnproductiveSymbols( const grammar::GNF & grammar );
template grammar::CNF TrimCFG::removeUnproductiveSymbols( const grammar::CNF & grammar );
template grammar::LG TrimCFG::removeUnproductiveSymbols( const grammar::LG & grammar );
template grammar::LeftLG TrimCFG::removeUnproductiveSymbols( const grammar::LeftLG & grammar );
template grammar::LeftRG TrimCFG::removeUnproductiveSymbols( const grammar::LeftRG & grammar );
template grammar::RightLG TrimCFG::removeUnproductiveSymbols( const grammar::RightLG & grammar );
template grammar::RightRG TrimCFG::removeUnproductiveSymbols( const grammar::RightRG & grammar );

template<class T>
T TrimCFG::trim( const T & grammar ) {
	return removeUnreachableSymbols( removeUnproductiveSymbols( grammar ) );
}

template grammar::CFG TrimCFG::trim( const grammar::CFG & grammar );
template grammar::EpsilonFreeCFG TrimCFG::trim( const grammar::EpsilonFreeCFG & grammar );
template grammar::GNF TrimCFG::trim( const grammar::GNF & grammar );
template grammar::CNF TrimCFG::trim( const grammar::CNF & grammar );
template grammar::LG TrimCFG::trim( const grammar::LG & grammar );
template grammar::LeftLG TrimCFG::trim( const grammar::LeftLG & grammar );
template grammar::LeftRG TrimCFG::trim( const grammar::LeftRG & grammar );
template grammar::RightLG TrimCFG::trim( const grammar::RightLG & grammar );
template grammar::RightRG TrimCFG::trim( const grammar::RightRG & grammar );

}

