/*
 * EpsilonNFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "EpsilonNFA.h"
#include "../../std/map.hpp"
#include "../AutomatonException.h"
#include <ostream>
#include <sstream>

namespace automaton {

AutomatonBase* EpsilonNFA::clone() const {
	return new EpsilonNFA(*this);
}

AutomatonBase* EpsilonNFA::plunder() && {
	return new EpsilonNFA(std::move(*this));
}

bool EpsilonNFA::removeState(const State& state) {
	if (initialStates.find(state) != initialStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >::const_iterator t = transitions.begin(); t != transitions.end(); t++) {
		if (t->first.first == state || t->second.find(state) != t->second.end())
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool EpsilonNFA::removeInputSymbol(const alphabet::Symbol& symbol) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(symbol);
	for (std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->first.second.is<alphabet::Symbol>() && transition->first.second.get<alphabet::Symbol>() == symbol)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used.");
	}

	return inputAlphabet.erase(symbol);

}

bool EpsilonNFA::addTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const State& to) {
	if (states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");

	if (input.is<alphabet::Symbol>() && inputAlphabet.find(input.get<alphabet::Symbol>()) == inputAlphabet.end())
		throw AutomatonException("Input symbol \"" + (std::string) input.get<alphabet::Symbol>() + "\" doesn't exist.");

	if (states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");

	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> > key = std::make_pair(from, input);
	
	return transitions[key].insert(to).second;
}

bool EpsilonNFA::addTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return addTransition(from, inputVariant, to);
}

bool EpsilonNFA::addTransition(const State& from, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return addTransition(from, inputVariant, to);
}

bool EpsilonNFA::removeTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> > key = std::make_pair(from, inputVariant);
	
	return transitions[key].erase(to);
}

bool EpsilonNFA::removeTransition(const State& from, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> > key = std::make_pair(from, inputVariant);
	
	return transitions[key].erase(to);
}

const std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& EpsilonNFA::getTransitions() const {
	return transitions;
}

std::map<State, std::set<State> > EpsilonNFA::getEpsilonTransitions() const {
	std::map<State, std::set<State> > result;
	for (std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->first.second.is<string::Epsilon>())
			result[transition->first.first].insert(transition->second.begin(), transition->second.end());
	}
	return result;
}

std::map<std::pair<State, alphabet::Symbol>, std::set<State> > EpsilonNFA::getSymbolTransitions() const {
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > result;
	for (std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->first.second.is<alphabet::Symbol>())
			result[std::pair<State, alphabet::Symbol>(transition->first.first, transition->first.second.get<alphabet::Symbol>())].insert(transition->second.begin(), transition->second.end());
	}
	return result;
}

std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> > EpsilonNFA::getTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> > transitionsFromState;
	for (std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->first.first == from) {
			transitionsFromState[transition->first].insert(transition->second.begin(), transition->second.end());
		}
	}

	return transitionsFromState;
}

std::map<State, std::set<State> > EpsilonNFA::getEpsilonTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> > key(from, std::variant<string::Epsilon, alphabet::Symbol>(string::Epsilon::EPSILON));
	std::map<State, std::set<State>> res;
	auto transition = transitions.find(key);
	if(transition != transitions.end())
		res.insert(std::make_pair(from, transition->second));
	return res;
}

std::map<std::pair<State, alphabet::Symbol>, std::set<State> > EpsilonNFA::getSymbolTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > transitionsFromState;
	for (std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->first.first == from && transition->first.second.is<alphabet::Symbol>()) {
			transitionsFromState[std::pair<State, alphabet::Symbol>(transition->first.first, transition->first.second.get<alphabet::Symbol>())].insert(transition->second.begin(), transition->second.end());
		}
	}

	return transitionsFromState;
}

std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> > EpsilonNFA::getTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> > transitionsToState;
	for (std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->second.find(to) != transition->second.end()) {
			transitionsToState[transition->first].insert(transition->second.begin(), transition->second.end());
		}
	}

	return transitionsToState;
}

std::map<State, std::set<State> > EpsilonNFA::getEpsilonTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<State, std::set<State> > transitionsToState;
	for (std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->second.find(to) != transition->second.end() && transition->first.second.is<string::Epsilon>()) {
			transitionsToState[transition->first.first].insert(transition->second.begin(), transition->second.end());
		}
	}

	return transitionsToState;
}

std::map<std::pair<State, alphabet::Symbol>, std::set<State> > EpsilonNFA::getSymbolTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > transitionsToState;
	for (std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->second.find(to) != transition->second.end() && transition->first.second.is<alphabet::Symbol>()) {
			transitionsToState[std::pair<State, alphabet::Symbol>(transition->first.first, transition->first.second.get<alphabet::Symbol>())].insert(transition->second.begin(), transition->second.end());
		}
	}

	return transitionsToState;
}

bool EpsilonNFA::isEpsilonFree() const {
	for (std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->first.second.is<string::Epsilon>()) {
			return false;
		}
	}
	return true;
}

bool EpsilonNFA::isDeterministic() const {
	if (initialStates.size() != 1 || !isEpsilonFree()) {
		return false;
	}

	for (std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->second.size() != 1 || transition->second.size() != 0) {
			return false;
		}
	}

	return true;
}

bool EpsilonNFA::isTotal() const {
	return isDeterministic() && transitions.size() == inputAlphabet.size() * states.size();
}

bool EpsilonNFA::operator==(const ObjectBase& other) const {
	return other == *this;
}

bool EpsilonNFA::operator<(const ObjectBase& other) const {
	return other > *this;
}

bool EpsilonNFA::operator>(const ObjectBase& other) const {
	return other < *this;
}

bool EpsilonNFA::operator==(const EpsilonNFA& other) const {
	return this->states == other.states && this->inputAlphabet == other.inputAlphabet && this->initialStates == other.initialStates && this->finalStates == other.finalStates && this->transitions == other.transitions;
}

bool EpsilonNFA::operator<(const EpsilonNFA& other) const {
	return std::tie(states, inputAlphabet, initialStates, finalStates, transitions) < std::tie(other.states, other.inputAlphabet, other.initialStates, other.finalStates, other.transitions);
}

void EpsilonNFA::operator>>(std::ostream& out) const {
	out << "(EpsilonNFA"
		<< " states = " << states
		<< " inputAlphabet = " << inputAlphabet
		<< " initialStates = " << initialStates
		<< " finalStates = " << finalStates
		<< " transitions = " << transitions
		<< ")";
}

EpsilonNFA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace automaton */
