/*
 * RealTimeHeightDeterministicNPDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#include "RealTimeHeightDeterministicNPDA.h"
#include "../../std/map.hpp"
#include "../AutomatonException.h"
#include <algorithm>
#include <sstream>

namespace automaton {

RealTimeHeightDeterministicNPDA::RealTimeHeightDeterministicNPDA(const alphabet::Symbol& bottomOfTheStackSymbol) : BottomOfTheStackSymbolPushdownStoreAlphabet(bottomOfTheStackSymbol) {

}

AutomatonBase* RealTimeHeightDeterministicNPDA::clone() const {
	return new RealTimeHeightDeterministicNPDA(*this);
}

AutomatonBase* RealTimeHeightDeterministicNPDA::plunder() && {
	return new RealTimeHeightDeterministicNPDA(std::move(*this));
}

bool RealTimeHeightDeterministicNPDA::removeState(const State& state) {
	if (initialStates.find(state) != initialStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::set<std::pair<State, alphabet::Symbol> > >& callTransition : callTransitions) {
		if (state == callTransition.first.first)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		for(const std::pair<State, alphabet::Symbol>& target : callTransition.second) {
			if(target.first == state)
				throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		}
	}

	for (const std::pair<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::set<State> >& returnTransition : returnTransitions) {
		if (state == std::get<0>(returnTransition.first))
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		for(const State& target : returnTransition.second) {
			if(target == state)
				throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		}
	}

	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::set<State> >& localTransition : localTransitions) {
		if (state == localTransition.first.first)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		for(const State& target : localTransition.second) {
			if(target == state)
				throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		}
	}

	return states.erase(state);
}

bool RealTimeHeightDeterministicNPDA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::set<std::pair<State, alphabet::Symbol> > >& callTransition : callTransitions) {
		if (callTransition.first.second.is<alphabet::Symbol>() && symbol == callTransition.first.second.get<alphabet::Symbol>())
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::set<State> >& returnTransition : returnTransitions) {
		if (std::get<1>(returnTransition.first).is<alphabet::Symbol>() && symbol == std::get<1>(returnTransition.first).get<alphabet::Symbol>())
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::set<State> >& localTransition : localTransitions) {
		if (localTransition.first.second.is<alphabet::Symbol>() && symbol == localTransition.first.second.get<alphabet::Symbol>())
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	return inputAlphabet.erase(symbol);
}

bool RealTimeHeightDeterministicNPDA::removeStackSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::set<std::pair<State, alphabet::Symbol> > >& callTransition : callTransitions) {
		for(const std::pair<State, alphabet::Symbol>& to : callTransition.second) {
			if (symbol == to.second)
				throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
		}
	}

	for (const std::pair<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::set<State> >& returnTransition : returnTransitions) {
		if (symbol == std::get<2>(returnTransition.first))
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	if(bottomOfTheStackSymbol == symbol) {
		throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is start symbol.");
	}

	return stackAlphabet.erase(symbol);
}

bool RealTimeHeightDeterministicNPDA::addCallTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const State& to, const alphabet::Symbol& push) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (input.is<alphabet::Symbol>() && inputAlphabet.find(input.get<alphabet::Symbol>()) == inputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input.get<alphabet::Symbol>() + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (stackAlphabet.find(push) == stackAlphabet.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) push + "\" doesn't exist.");
	}

	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>> key(from, input);
	std::pair<automaton::State, alphabet::Symbol> value = std::make_pair(to, push);
	
	return callTransitions[key].insert(value).second;
}

bool RealTimeHeightDeterministicNPDA::addCallTransition(const State& from, const State& to, const alphabet::Symbol& push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return addCallTransition(from, inputVariant, to, push);
}

bool RealTimeHeightDeterministicNPDA::addCallTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return addCallTransition(from, inputVariant, to, push);
}

bool RealTimeHeightDeterministicNPDA::addReturnTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const alphabet::Symbol& pop, const State& to) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (input.is<alphabet::Symbol>() && inputAlphabet.find(input.get<alphabet::Symbol>()) == inputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input.get<alphabet::Symbol>() + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (stackAlphabet.find(pop) == stackAlphabet.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) pop + "\" doesn't exist.");
	}

	std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol> key(from, input, pop);

	return returnTransitions[key].insert(to).second;
}

bool RealTimeHeightDeterministicNPDA::addReturnTransition(const State& from, const alphabet::Symbol& pop, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return addReturnTransition(from, inputVariant, pop, to);
}

bool RealTimeHeightDeterministicNPDA::addReturnTransition(const State& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return addReturnTransition(from, inputVariant, pop, to);
}

bool RealTimeHeightDeterministicNPDA::addLocalTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const State& to) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (input.is<alphabet::Symbol>() && inputAlphabet.find(input.get<alphabet::Symbol>()) == inputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input.get<alphabet::Symbol>() + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>> key(from, input);

	return localTransitions[key].insert(to).second;
}

bool RealTimeHeightDeterministicNPDA::addLocalTransition(const State& from, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return addLocalTransition(from, inputVariant, to);
}

bool RealTimeHeightDeterministicNPDA::addLocalTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return addLocalTransition(from, inputVariant, to);
}

bool RealTimeHeightDeterministicNPDA::removeCallTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const State& to, const alphabet::Symbol& push) {
	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>> key(from, input);
	std::pair<automaton::State, alphabet::Symbol> value = std::make_pair(to, push);
	
	return callTransitions[key].erase(value);
}

bool RealTimeHeightDeterministicNPDA::removeCallTransition(const State& from, const State& to, const alphabet::Symbol& push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return removeCallTransition(from, inputVariant, to, push);
}

bool RealTimeHeightDeterministicNPDA::removeCallTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return removeCallTransition(from, inputVariant, to, push);
}

bool RealTimeHeightDeterministicNPDA::removeReturnTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const alphabet::Symbol& pop, const State& to) {
	std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol> key(from, input, pop);

	return returnTransitions[key].erase(to);
}

bool RealTimeHeightDeterministicNPDA::removeReturnTransition(const State& from, const alphabet::Symbol& pop, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return removeReturnTransition(from, inputVariant, pop, to);
}

bool RealTimeHeightDeterministicNPDA::removeReturnTransition(const State& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return removeReturnTransition(from, inputVariant, pop, to);
}

bool RealTimeHeightDeterministicNPDA::removeLocalTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const State& to) {
	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>> key(from, input);

	return localTransitions[key].erase(to);
}

bool RealTimeHeightDeterministicNPDA::removeLocalTransition(const State& from, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return removeLocalTransition(from, inputVariant, to);
}

bool RealTimeHeightDeterministicNPDA::removeLocalTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return removeLocalTransition(from, inputVariant, to);
}

const std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::set<std::pair<State, alphabet::Symbol> > >& RealTimeHeightDeterministicNPDA::getCallTransitions() const {
		return callTransitions;
}

const std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::set<State> >& RealTimeHeightDeterministicNPDA::getReturnTransitions() const {
		return returnTransitions;
}

const std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::set<State> >& RealTimeHeightDeterministicNPDA::getLocalTransitions() const {
		return localTransitions;
}

bool RealTimeHeightDeterministicNPDA::operator==(const ObjectBase& other) const {
	return other == *this;
}

bool RealTimeHeightDeterministicNPDA::operator<(const ObjectBase& other) const {
	return other > *this;
}

bool RealTimeHeightDeterministicNPDA::operator>(const ObjectBase& other) const {
	return other < *this;
}

bool RealTimeHeightDeterministicNPDA::operator==(const RealTimeHeightDeterministicNPDA& other) const {
	return this->states == other.states && this->inputAlphabet == other.inputAlphabet && this->initialStates == other.initialStates && this->finalStates == other.finalStates && this->stackAlphabet == other.stackAlphabet && this->bottomOfTheStackSymbol == other.bottomOfTheStackSymbol && this->callTransitions == other.callTransitions && this->returnTransitions == other.returnTransitions && this->localTransitions == other.localTransitions;
}

bool RealTimeHeightDeterministicNPDA::operator<(const RealTimeHeightDeterministicNPDA& other) const {
	return std::tie(states, inputAlphabet, initialStates, finalStates, stackAlphabet, bottomOfTheStackSymbol, callTransitions, returnTransitions, localTransitions) < std::tie(other.states, other.inputAlphabet, other.initialStates, other.finalStates, other.stackAlphabet, other.bottomOfTheStackSymbol, other.callTransitions, other.returnTransitions, other.localTransitions);
}

void RealTimeHeightDeterministicNPDA::operator>>(std::ostream& out) const {
	out << "(RealTimeHeightDeterministicNPDA"
		<< "states = " << states
		<< "inputAlphabet = " << inputAlphabet
		<< "initialStates = " << initialStates
		<< "finalStates = " << finalStates
		<< "stackAlphabet = " << stackAlphabet
		<< "bottomOfTheStackSymbol = " << bottomOfTheStackSymbol
		<< "callTransitions = " << callTransitions
		<< "returnTransitions = " << returnTransitions
		<< "localTransitions = " << localTransitions
		<< ")";
}

RealTimeHeightDeterministicNPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace automaton */
