/*
 * UnknownTransition.cpp
 *
 *  Created on: Oct 13, 2013
 *      Author: Martin Zak
 */

#include "UnknownTransition.h"
#include "../std/vector.hpp"
#include "../std/pointer.hpp"

namespace automaton {

UnknownTransition::UnknownTransition() : from(NULL), to(NULL), input(NULL), output(NULL), shift(NOT_SET) {

}

UnknownTransition::UnknownTransition(const State& from, const State& to, const std::vector<alphabet::Symbol>& pop, const std::vector<alphabet::Symbol>& push, const std::variant<string::Epsilon, alphabet::Symbol>& input, const std::variant<string::Epsilon, alphabet::Symbol>& output, Shift shift) :
		from(new State(from)), to(new State(to)), pop(pop), push(push), input(new std::variant<string::Epsilon, alphabet::Symbol>(input)), output(new std::variant<string::Epsilon, alphabet::Symbol>(output)), shift(shift) {

}

UnknownTransition::UnknownTransition(const UnknownTransition& other) :
		from(NULL), to(NULL), pop(other.pop), push(other.push), input(NULL), output(NULL), shift(other.shift) {
	if(other.from)
		from = new State(*other.from);
	if(other.to)
		to = new State(*other.to);
	if(other.input)
		input = new std::variant<string::Epsilon, alphabet::Symbol>(*other.input);
	if(other.output)
		output = new std::variant<string::Epsilon, alphabet::Symbol>(*other.output);
}

UnknownTransition::UnknownTransition(UnknownTransition&& other) noexcept :
		from(NULL), to(NULL), pop(std::move(other.pop)), push(std::move(other.push)), input(NULL), output(NULL), shift(std::move(other.shift)) {
	if(other.from)
		from = new State(std::move(*other.from));
	if(other.to)
		to = new State(std::move(*other.to));
	if(other.input)
		input = new std::variant<string::Epsilon, alphabet::Symbol>(std::move(*other.input));
	if(other.output)
		output = new std::variant<string::Epsilon, alphabet::Symbol>(std::move(*other.output));
}

const UnknownTransition& UnknownTransition::operator=(const UnknownTransition& other) {
	UnknownTransition tmp(other);

	std::swap(from, tmp.from);
	std::swap(to, tmp.to);
	std::swap(pop, tmp.pop);
	std::swap(push, tmp.push);
	std::swap(input, tmp.input);
	std::swap(output, tmp.output);
	std::swap(shift, tmp.shift);

	return *this;
}

const UnknownTransition& UnknownTransition::operator=(UnknownTransition&& other) noexcept {
	std::swap(from, other.from);
	std::swap(to, other.to);
	std::swap(pop, other.pop);
	std::swap(push, other.push);
	std::swap(input, other.input);
	std::swap(output, other.output);
	std::swap(shift, other.shift);

	return *this;
}

UnknownTransition::~UnknownTransition() noexcept {
	delete from;
	delete to;
}

void UnknownTransition::setFrom(const State& state) {
	delete from;
	from = new State(state);
}

bool UnknownTransition::hasFrom() const {
	return from != NULL;
}

void UnknownTransition::clearFrom() {
	delete from;
	from = NULL;
}

const State& UnknownTransition::getFrom() const {
	return *from;
}

void UnknownTransition::setTo(const State& state) {
	delete to;
	to = new State(state);
}

bool UnknownTransition::hasTo() const {
	return to != NULL;
}

void UnknownTransition::clearTo() {
	delete to;
	to = NULL;
}

const State& UnknownTransition::getTo() const {
	return *to;
}

void UnknownTransition::setPop(const std::vector<alphabet::Symbol>& pop) {
	this->pop = pop;
}

const std::vector<alphabet::Symbol>& UnknownTransition::getPop() const {
	return pop;
}

void UnknownTransition::addPop(const alphabet::Symbol& symbol) {
	pop.push_back(symbol);
}

void UnknownTransition::setPush(const std::vector<alphabet::Symbol>& push) {
	this->push = push;
}

const std::vector<alphabet::Symbol>& UnknownTransition::getPush() const {
	return push;
}

void UnknownTransition::addPush(const alphabet::Symbol& symbol) {
	push.push_back(symbol);
}

void UnknownTransition::setOutput(const std::variant<string::Epsilon, alphabet::Symbol>& symbol) {
	delete output;
	output = new std::variant<string::Epsilon, alphabet::Symbol>(symbol);
}

bool UnknownTransition::hasOutput() const {
	return output != NULL;
}

void UnknownTransition::clearOutput() {
	delete output;
	output = NULL;
}

const std::variant<string::Epsilon, alphabet::Symbol>& UnknownTransition::getOutput() const {
	return *output;
}

void UnknownTransition::setInput(const std::variant<string::Epsilon, alphabet::Symbol>& symbol) {
	delete input;
	input = new std::variant<string::Epsilon, alphabet::Symbol>(symbol);
}

bool UnknownTransition::hasInput() const {
	return input != NULL;
}

void UnknownTransition::clearInput() {
	delete input;
	input = NULL;
}

const std::variant<string::Epsilon, alphabet::Symbol>& UnknownTransition::getInput() const {
	return *input;
}

const Shift& UnknownTransition::getShift() const {
	return shift;
}

void UnknownTransition::setShift(const Shift& shift) {
	this->shift = shift;
}

bool UnknownTransition::operator <(const UnknownTransition& other) const {
	std::pointer<automaton::State> fromPointer(from);
	std::pointer<automaton::State> otherFromPointer(other.from);

	std::pointer<std::variant<string::Epsilon, alphabet::Symbol>> inputPointer(input);
	std::pointer<std::variant<string::Epsilon, alphabet::Symbol>> otherInputPointer(other.input);

	std::pointer<automaton::State> toPointer(to);
	std::pointer<automaton::State> otherToPointer(other.to);

	std::pointer<std::variant<string::Epsilon, alphabet::Symbol>> outputPointer(output);
	std::pointer<std::variant<string::Epsilon, alphabet::Symbol>> otherOutputPointer(other.output);

	return std::tie(fromPointer, inputPointer, pop, shift, toPointer, outputPointer, push) < std::tie(otherFromPointer, otherInputPointer, pop, shift, otherToPointer, otherOutputPointer, push);
}

bool UnknownTransition::operator ==(const UnknownTransition& other) const {
	std::pointer<automaton::State> fromPointer(from);
	std::pointer<automaton::State> otherFromPointer(other.from);

	std::pointer<std::variant<string::Epsilon, alphabet::Symbol>> inputPointer(input);
	std::pointer<std::variant<string::Epsilon, alphabet::Symbol>> otherInputPointer(other.input);

	std::pointer<automaton::State> toPointer(to);
	std::pointer<automaton::State> otherToPointer(other.to);

	std::pointer<std::variant<string::Epsilon, alphabet::Symbol>> outputPointer(output);
	std::pointer<std::variant<string::Epsilon, alphabet::Symbol>> otherOutputPointer(other.output);

	return fromPointer == otherFromPointer && inputPointer == otherInputPointer && pop == other.pop && shift == other.shift && toPointer == otherToPointer && outputPointer == otherOutputPointer && push == other.push;
}

bool UnknownTransition::operator !=(const UnknownTransition& other) const {
	return !(*this == other);
}

std::ostream& operator<<(std::ostream& out, const UnknownTransition& transition) {
	std::pointer<automaton::State> fromPointer(transition.from);
	std::pointer<std::variant<string::Epsilon, alphabet::Symbol>> inputPointer(transition.input);
	std::pointer<automaton::State> toPointer(transition.to);
	std::pointer<std::variant<string::Epsilon, alphabet::Symbol>> outputPointer(transition.output);

	out << "(UnknownTransition"
		<< " from = " << fromPointer
		<< " input = " << inputPointer
		<< " to = " << toPointer
		<< " output = " << outputPointer
		<< " pop = " << transition.pop
		<< " push = " << transition.push
		<< " shift = " << SHIFT_NAMES[transition.shift]
		<< ")";

	return out;
}

} /* namespace automaton */
