/*
 * AutomatonConvertor.cpp
 *
 *  Created on: Apr 23, 2013
 *      Author: Jan Travnicek
 */

#include "AutomatonConvertor.h"

#include <algorithm>
#include <iostream>

#include "../automaton/AutomatonException.h"

namespace automaton {

Automaton AutomatonConvertor::buildAutomaton(const UnknownAutomaton& automaton) {
	if (isEpsilonNFA(automaton)) {
		return Automaton(buildEpsilonNFA(automaton));
	} else if (isNPDA(automaton)) {
		return Automaton(buildNPDA(automaton));
	} else if (isOneTapeDTM(automaton)) {
		return Automaton(buildOneTapeDTM(automaton));
	} else {
		throw AutomatonException("Cannot determine automaton type.");
	}
}

bool AutomatonConvertor::isEpsilonNFA(const UnknownAutomaton& automaton) {
	if (automaton.getStackSymbols().size() > 0 || automaton.getInitialSymbols().size() > 0
			|| automaton.getTapeSymbols().size() > 0 || automaton.hasBlankSymbol()) {
		return false;
	}

	const std::set<UnknownTransition> transitions = automaton.getTransitions();
	std::set<UnknownTransition>::const_iterator it = transitions.begin();
	while (it != transitions.end()) {
		if (!isEpsilonNFATransition(*it)) {
			return false;
		}
		it++;
	}

	return true;
}

bool AutomatonConvertor::isEpsilonNFATransition(const UnknownTransition& transition) {
	return !(transition.getPop().size() > 0 || transition.getPush().size() > 0 || transition.hasOutput()
			|| transition.getShift() != NOT_SET);
}

EpsilonNFA AutomatonConvertor::buildEpsilonNFA(const UnknownAutomaton& automaton) {
	EpsilonNFA fsm;

	fsm.setStates(automaton.getStates());
	fsm.setInputSymbols(automaton.getInputSymbols());
	fsm.setInitialStates(automaton.getInitialStates());
	fsm.setFinalStates(automaton.getFinalStates());

	const std::set<UnknownTransition> transitions = automaton.getTransitions();
	std::set<UnknownTransition>::const_iterator transition = transitions.begin();
	while (transition != transitions.end()) {
		fsm.addTransition(transition->getFrom(), transition->getInput(), transition->getTo());
		transition++;
	}

	return fsm;
}

bool AutomatonConvertor::isNPDA(const UnknownAutomaton& automaton) {
	if (automaton.getTapeSymbols().size() > 0 || automaton.hasBlankSymbol()) {
		return false;
	}

	const std::set<UnknownTransition> transitions = automaton.getTransitions();
	std::set<UnknownTransition>::const_iterator transition = transitions.begin();
	while (transition != transitions.end()) {
		if (!isNPDATransition(*transition)) {
			return false;
		}
		transition++;
	}

	return true;
}

bool AutomatonConvertor::isNPDATransition(const UnknownTransition& transition) {
	return !(transition.hasOutput() || transition.getShift() != NOT_SET);
}

NPDA AutomatonConvertor::buildNPDA(const UnknownAutomaton& automaton) {
	NPDA pda;
	pda.setStates(automaton.getStates());
	pda.setInputSymbols(automaton.getInputSymbols());
	pda.setInitialStates(automaton.getInitialStates());
	pda.setFinalStates(automaton.getFinalStates());
	pda.setStackSymbols(automaton.getStackSymbols());
	pda.setInitialSymbols(automaton.getInitialSymbols());

	const std::set<UnknownTransition>& transitions = automaton.getTransitions();
	std::set<UnknownTransition>::const_iterator transition = transitions.begin();
	while (transition != transitions.end()) {
		pda.addTransition(transition->getFrom(), transition->getInput().get<alphabet::Symbol>(), transition->getPop(), transition->getTo(), transition->getPush());
		transition++;
	}

	return pda;
}

bool AutomatonConvertor::isOneTapeDTM(const UnknownAutomaton& automaton) {
	if(!automaton.hasBlankSymbol()) return false;

	if(automaton.getInitialStates().size() > 1) return false;

	if (automaton.getStackSymbols().size() > 0 || automaton.getInitialSymbols().size() > 0) {
		return false;
	}

	const std::set<UnknownTransition>& transitions = automaton.getTransitions();
	std::set<UnknownTransition>::const_iterator transition = transitions.begin();
	while (transition != transitions.end()) {
		if (!isOneTapeDTMTransition(*transition)) {
			return false;
		}
		transition++;
	}

	return true;

}

bool AutomatonConvertor::isOneTapeDTMTransition(const UnknownTransition& transition) {
	return !(transition.getPop().size() > 0 || transition.getPush().size() > 0);

}

OneTapeDTM AutomatonConvertor::buildOneTapeDTM(const UnknownAutomaton& automaton) {
	OneTapeDTM tm(*automaton.getInitialStates().begin(), automaton.getBlankSymbol());
	
	tm.setTapeSymbols(automaton.getTapeSymbols());
	tm.setStates(automaton.getStates());
	tm.setInputSymbols(automaton.getInputSymbols());
	tm.setFinalStates(automaton.getFinalStates());

	const std::set<UnknownTransition>& transitions = automaton.getTransitions();
	std::set<UnknownTransition>::const_iterator transition = transitions.begin();
	while (transition != transitions.end()) {
		tm.addTransition(transition->getFrom(), transition->getInput().get<alphabet::Symbol>(), transition->getTo(), transition->getOutput().get<alphabet::Symbol>(), transition->getShift());
		transition++;
	}

	return tm;

}

} /* namespace automaton */
