/*
 * DataFactory.hpp
 *
 *  Created on: Jan 1, 2014
 *      Author: Jan Travnicek
 */

#ifndef DATA_FACTORY_H_
#define DATA_FACTORY_H_

#include <string>
#include <list>

#include "../sax/Token.h"
#include "../sax/SaxParseInterface.h"
#include "../sax/SaxComposeInterface.h"
#include "../Api.hpp"

namespace alib {

/**
 * Data builder.
 */
class DataFactory {
public:
	/**
	 * Parses the XML in file and returns the String.
	 * @param filename path to the file
	 * @return String
	 */
	template<class T>
	static T fromFile(const std::string& filename) {
		std::list<sax::Token> tokens;
		sax::SaxParseInterface::parseFile(filename, tokens);
		return fromTokens<T>(tokens);
	}

	/**
	 * Parses the XML and returns the String.
	 * @param str string containing the XML
	 * @return String
	 */
	template<class T>
	static T fromString(const std::string& str) {
		std::list<sax::Token> tokens;
		sax::SaxParseInterface::parseMemory(str, tokens);
		return fromTokens<T>(tokens);
	}
	
	/**
	 * Parses the XML from stdin and returns the String.
	 * @return String
	 */
	template<class T>
	static T fromStdin() {
		return fromStream<T>(std::cin);
	}
	
	/**
	 * Parses the XML from stream and returns the String.
	 * @return String
	 */
	template<class T>
	static T fromStream(std::istream& in) {
		std::string input{std::istreambuf_iterator<char>(in), std::istreambuf_iterator<char>()};
		return fromString<T>(input);
	}

	/**
	 * Parses the String from list of tokens.
	 * @param tokens XML represented as list of tokens
	 * @return parsed String
	 */
	template<class T>
	static T fromTokens(std::list<sax::Token> tokens) {
		T res = alib::api<T>::parse(tokens);
		if(tokens.size() != 0) throw exception::AlibException("Unexpeted tokens at the end of the xml");
		return res;
	}

	/**
	 * Parses the XML in file and returns the String.
	 * @param filename path to the file
	 * @return String
	 */
	template<class T>
	static void toFile(const T& data, const std::string& filename) {
		std::list<sax::Token> tokens = toTokens<T>(data);
		sax::SaxComposeInterface::printFile(filename, tokens);
	}

	/**
	 * Parses the XML and returns the String.
	 * @param str string containing the XML
	 * @return String
	 */
	template<class T>
	static std::string toString(const T& data) {
		std::list<sax::Token> tokens = toTokens<T>(data);
		std::string str;
		sax::SaxComposeInterface::printMemory(str, tokens);
		return str;
	}
	
	/**
	 * Parses the XML from stdin and returns the String.
	 * @return String
	 */
	template<class T>
	static void toStdout(const T& data) {
		return toStream<T>(data, std::cout);
	}
	
	/**
	 * Parses the XML from stream and returns the String.
	 * @return String
	 */
	template<class T>
	static void toStream(const T& data, std::ostream& out) {
		std::list<sax::Token> tokens = toTokens<T>(data);
		sax::SaxComposeInterface::printStream(out, tokens);
	}

	/**
	 * Parses the String from list of tokens.
	 * @param tokens XML represented as list of tokens
	 * @return parsed String
	 */
	template<class T>
	static std::list<sax::Token> toTokens(const T& data) {
		return alib::api<T>::compose(data);
	}

};

} /* namespace alib */

#endif /* DATA_FACTORY_H_ */
