/*
 * RegExpToXMLComposer.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "RegExpToXMLComposer.h"

#include "../Api.hpp"

namespace regexp {

void RegExpToXMLComposer::Visit(void* userData, const UnboundedRegExpAlternation& alternation) const {
	std::list<sax::Token> &out = *((std::list<sax::Token>*) userData);

	out.push_back(sax::Token("alternation", sax::Token::TokenType::START_ELEMENT));
	for (const auto& element : alternation.getElements()) {
		const UnboundedRegExpElement& object = static_cast<const UnboundedRegExpElement&>(*element);
		object.Accept(userData, *this);
	}
	out.push_back(sax::Token("alternation", sax::Token::TokenType::END_ELEMENT));
}

void RegExpToXMLComposer::Visit(void* userData, const UnboundedRegExpConcatenation& concatenation) const {
	std::list<sax::Token> &out = *((std::list<sax::Token>*) userData);

	out.push_back(sax::Token("concatenation", sax::Token::TokenType::START_ELEMENT));
	for (auto element : concatenation.getElements()) {
		const UnboundedRegExpElement& object = static_cast<const UnboundedRegExpElement&>(*element);
		object.Accept(userData, *this);
	}
	out.push_back(sax::Token("concatenation", sax::Token::TokenType::END_ELEMENT));

}

void RegExpToXMLComposer::Visit(void* userData, const UnboundedRegExpIteration& iteration) const {
	std::list<sax::Token> &out = *((std::list<sax::Token>*) userData);

	out.push_back(sax::Token("iteration", sax::Token::TokenType::START_ELEMENT));
	const UnboundedRegExpElement& object = static_cast<const UnboundedRegExpElement&>(iteration.getElement());
	object.Accept(userData, *this);
	out.push_back(sax::Token("iteration", sax::Token::TokenType::END_ELEMENT));
}

void RegExpToXMLComposer::Visit(void* userData, const UnboundedRegExpSymbol& symbol) const {
	std::list<sax::Token> &out = *((std::list<sax::Token>*) userData);

	out.splice(out.end(), alib::api<alphabet::Symbol>::compose(symbol.getSymbol()));
}

void RegExpToXMLComposer::Visit(void* userData, const UnboundedRegExpEpsilon&) const {
	std::list<sax::Token> &out = *((std::list<sax::Token>*) userData);

	out.push_back(sax::Token("epsilon", sax::Token::TokenType::START_ELEMENT));
	out.push_back(sax::Token("epsilon", sax::Token::TokenType::END_ELEMENT));
}

void RegExpToXMLComposer::Visit(void* userData, const UnboundedRegExpEmpty&) const {
	std::list<sax::Token> &out = *((std::list<sax::Token>*) userData);

	out.push_back(sax::Token("empty", sax::Token::TokenType::START_ELEMENT));
	out.push_back(sax::Token("empty", sax::Token::TokenType::END_ELEMENT));
}

void RegExpToXMLComposer::Visit(void* userData, const FormalRegExpAlternation& alternation) const {
	std::list<sax::Token> &out = *((std::list<sax::Token>*) userData);

	out.push_back(sax::Token("alternation", sax::Token::TokenType::START_ELEMENT));
	{
		const auto& element = alternation.getLeftElement();
		const FormalRegExpElement& object = static_cast<const FormalRegExpElement&>(element);
		object.Accept(userData, *this);
	}
	{
		const auto& element = alternation.getRightElement();
		const FormalRegExpElement& object = static_cast<const FormalRegExpElement&>(element);
		object.Accept(userData, *this);
	}
	out.push_back(sax::Token("alternation", sax::Token::TokenType::END_ELEMENT));
}

void RegExpToXMLComposer::Visit(void* userData, const FormalRegExpConcatenation& concatenation) const {
	std::list<sax::Token> &out = *((std::list<sax::Token>*) userData);

	out.push_back(sax::Token("concatenation", sax::Token::TokenType::START_ELEMENT));
	{
		const auto& element = concatenation.getLeftElement();
		const FormalRegExpElement& object = static_cast<const FormalRegExpElement&>(element);
		object.Accept(userData, *this);
	}
	{
		const auto& element = concatenation.getRightElement();
		const FormalRegExpElement& object = static_cast<const FormalRegExpElement&>(element);
		object.Accept(userData, *this);
	}
	out.push_back(sax::Token("concatenation", sax::Token::TokenType::END_ELEMENT));

}

void RegExpToXMLComposer::Visit(void* userData, const FormalRegExpIteration& iteration) const {
	std::list<sax::Token> &out = *((std::list<sax::Token>*) userData);

	out.push_back(sax::Token("iteration", sax::Token::TokenType::START_ELEMENT));
	const FormalRegExpElement& object = static_cast<const FormalRegExpElement&>(iteration.getElement());
	object.Accept(userData, *this);
	out.push_back(sax::Token("iteration", sax::Token::TokenType::END_ELEMENT));
}

void RegExpToXMLComposer::Visit(void* userData, const FormalRegExpSymbol& symbol) const {
	std::list<sax::Token> &out = *((std::list<sax::Token>*) userData);

	out.splice(out.end(), alib::api<alphabet::Symbol>::compose(symbol.getSymbol()));
}

void RegExpToXMLComposer::Visit(void* userData, const FormalRegExpEpsilon&) const {
	std::list<sax::Token> &out = *((std::list<sax::Token>*) userData);

	out.push_back(sax::Token("epsilon", sax::Token::TokenType::START_ELEMENT));
	out.push_back(sax::Token("epsilon", sax::Token::TokenType::END_ELEMENT));
}

void RegExpToXMLComposer::Visit(void* userData, const FormalRegExpEmpty&) const {
	std::list<sax::Token> &out = *((std::list<sax::Token>*) userData);

	out.push_back(sax::Token("empty", sax::Token::TokenType::START_ELEMENT));
	out.push_back(sax::Token("empty", sax::Token::TokenType::END_ELEMENT));
}

std::list<sax::Token> RegExpToXMLComposer::compose(const RegExp& regexp) const {
	std::list<sax::Token> out;
	regexp.getData().Accept((void*) &out, alib::ToXMLComposers::toXMLComposers);
	return out;
}

std::list<sax::Token> RegExpToXMLComposer::compose(const RegExpBase& regexp) const {
	std::list<sax::Token> out;
	regexp.Accept((void*) &out, alib::ToXMLComposers::toXMLComposers);
	return out;
}

std::list<sax::Token> RegExpToXMLComposer::compose(const UnboundedRegExp& regexp) const {
	std::list<sax::Token> out;
	out.push_back(sax::Token("unboundedRegexp", sax::Token::TokenType::START_ELEMENT));
	{
		out.push_back(sax::Token("alphabet", sax::Token::TokenType::START_ELEMENT));
		for (const auto& symbol: regexp.getAlphabet()) {
			out.splice(out.end(), alib::api<alphabet::Symbol>::compose(symbol));
		}
		out.push_back(sax::Token("alphabet", sax::Token::TokenType::END_ELEMENT));
	}
	regexp.getRegExp().Accept((void*) &out, *this);
	out.push_back(sax::Token("unboundedRegexp", sax::Token::TokenType::END_ELEMENT));
	return out;
}

std::list<sax::Token> RegExpToXMLComposer::compose(const FormalRegExp& regexp) const {
	std::list<sax::Token> out;
	out.push_back(sax::Token("formalRegexp", sax::Token::TokenType::START_ELEMENT));
	{
		out.push_back(sax::Token("alphabet", sax::Token::TokenType::START_ELEMENT));
		for (const auto& symbol: regexp.getAlphabet()) {
			out.splice(out.end(), alib::api<alphabet::Symbol>::compose(symbol));
		}
		out.push_back(sax::Token("alphabet", sax::Token::TokenType::END_ELEMENT));
	}
	regexp.getRegExp().Accept((void*) &out, *this);
	out.push_back(sax::Token("formalRegexp", sax::Token::TokenType::END_ELEMENT));
	return out;
}

} /* namespace regexp */
